# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with GraphDati.

Note that these are not exact I/O - you can't currently use them as a round trip because
the input functions expect the GraphDati format that's output by BioDati.
"""

import gzip
import json
import logging
from collections import defaultdict
from typing import Any, Iterable, List, Mapping, Optional, TextIO, Tuple, Union

import pyparsing
from networkx.utils import open_file
from tqdm.autonotebook import tqdm

from .jgif import NAMESPACE_TO_PATTERN
from ..canonicalize import edge_to_tuple
from ..constants import (
    ANNOTATIONS,
    CITATION,
    CITATION_TYPE_PUBMED,
    CITATION_TYPE_URL,
    EVIDENCE,
    IDENTIFIER,
    NAMESPACE,
    RELATION,
    UNQUALIFIED_EDGES,
)
from ..parser import BELParser
from ..struct import BELGraph
from ..typing import EdgeData

__all__ = [
    "to_graphdati",
    "from_graphdati",
    "to_graphdati_file",
    "from_graphdati_file",
    "to_graphdati_gz",
    "from_graphdati_gz",
    "to_graphdati_jsons",
    "from_graphdati_jsons",
    "to_graphdati_jsonl",
    "to_graphdati_jsonl_gz",
]

logger = logging.getLogger(__name__)

NanopubMapping = Mapping[str, Mapping[str, Any]]

SCHEMA_URI = "https://github.com/belbio/schemas/blob/master/schemas/nanopub_bel-1.0.0.yaml"
GRAPHDATI_PUBLICATION_TYPES = {
    "PMID": CITATION_TYPE_PUBMED,
    "http": CITATION_TYPE_URL,
    "https": CITATION_TYPE_URL,
}


@open_file(1, mode="w")
def to_graphdati_file(graph: BELGraph, path: Union[str, TextIO], use_identifiers: bool = True, **kwargs) -> None:
    """Write this graph as GraphDati JSON to a file.

    :param graph: A BEL graph
    :param path: A path or file-like
    """
    json.dump(to_graphdati(graph, use_identifiers=use_identifiers), path, ensure_ascii=False, **kwargs)


def from_graphdati_file(path: Union[str, TextIO]) -> BELGraph:
    """Load a file containing GraphDati JSON.

    :param path: A path or file-like
    """
    return from_graphdati(json.load(path))


def to_graphdati_gz(graph: BELGraph, path: str, **kwargs) -> None:
    """Write a graph as GraphDati JSON to a gzip file."""
    with gzip.open(path, "wt") as file:
        to_graphdati_file(graph, file, **kwargs)


def from_graphdati_gz(path: str) -> BELGraph:
    """Read a graph as GraphDati JSON from a gzip file."""
    with gzip.open(path, "rt") as file:
        return from_graphdati(json.load(file))


def to_graphdati_jsons(graph: BELGraph, **kwargs) -> str:
    """Dump this graph as a GraphDati JSON object to a string.

    :param graph: A BEL graph
    """
    return json.dumps(to_graphdati(graph), ensure_ascii=False, **kwargs)


def from_graphdati_jsons(s: str) -> BELGraph:
    """Load a graph from a GraphDati JSON string.

    :param graph: A BEL graph
    """
    return from_graphdati(json.loads(s))


@open_file(1, mode="w")
def to_graphdati_jsonl(graph, file, use_identifiers: bool = True, use_tqdm: bool = True):
    """Write this graph as a GraphDati JSON lines file.

    :param graph: A BEL graph
    """
    for nanopub in _iter_graphdati(graph, use_identifiers=use_identifiers, use_tqdm=use_tqdm):
        print(json.dumps(nanopub), file=file)


def to_graphdati_jsonl_gz(graph: BELGraph, path: str, **kwargs) -> None:
    """Write a graph as GraphDati JSONL to a gzip file.

    :param graph: A BEL graph
    """
    with gzip.open(path, "wt") as file:
        to_graphdati_jsonl(graph, file, **kwargs)


def to_graphdati(
    graph,
    *,
    use_identifiers: bool = True,
    skip_unqualified: bool = True,
    use_tqdm: bool = False,
    metadata_extras: Optional[Mapping[str, Any]] = None,
) -> List[NanopubMapping]:
    """Export a GraphDati list using the nanopub.

    :param graph: A BEL graph
    :param use_identifiers: use OBO-style identifiers
    :param use_tqdm: Show a progress bar while generating nanopubs
    :param skip_unqualified: Should unqualified edges be output as nanopubs? Defaults to false.
    :param metadata_extras: Extra information to pass into the metadata part of nanopubs
    """
    return list(
        _iter_graphdati(
            graph,
            use_identifiers=use_identifiers,
            skip_unqualified=skip_unqualified,
            metadata_extras=metadata_extras,
            use_tqdm=use_tqdm,
        )
    )


def _iter_graphdati(
    graph,
    *,
    skip_unqualified: bool = True,
    use_identifiers: bool = True,
    use_tqdm: bool = False,
    metadata_extras: Optional[Mapping[str, Any]] = None,
) -> Iterable[NanopubMapping]:
    it = graph.edges(keys=True, data=True)
    if use_tqdm:
        it = tqdm(it, total=graph.number_of_edges(), desc="iterating as nanopubs")
    for u, v, k, d in it:
        if skip_unqualified and d[RELATION] in UNQUALIFIED_EDGES:
            continue
        yield _make_nanopub(graph, u, v, k, d, use_identifiers, metadata_extras=metadata_extras)


def _make_nanopub(graph: BELGraph, u, v, k, d, use_identifiers, metadata_extras=None) -> NanopubMapping:
    return dict(
        nanopub=dict(
            schema_uri=SCHEMA_URI,
            type=dict(name="BEL", version="2.1.0"),
            annotations=_get_annotations(d),
            citation=_get_citation(d),
            assertions=_get_assertions(u, v, d, use_identifiers),
            evidence=_get_evidence(d),
            metadata=_get_metadata(graph, d, extras=metadata_extras),
            id="pybel_{}".format(k),
        ),
    )


def _get_assertions(u, v, d, use_identifiers):
    return [
        dict(
            zip(
                ("subject", "relation", "object"),
                edge_to_tuple(u, v, d, use_identifiers=use_identifiers),
            )
        ),
    ]


def _get_evidence(d):
    return d.get(EVIDENCE, "Not Available")


def _get_citation(d):
    citation = d.get(CITATION)
    rv = {}
    if citation is None:
        rv["reference"] = "Not Available"
    else:
        rv["database"] = dict(name=citation[NAMESPACE], id=citation[IDENTIFIER])
    return rv


def _get_metadata(graph: BELGraph, _, extras=None):
    rv = dict(
        gd_creator=graph.authors,
        version=graph.version,
    )  # TODO later
    if extras is not None:
        rv.update(extras)
    return rv


def _get_annotations(d: EdgeData) -> List[Mapping[str, str]]:
    rv = []
    for key, values in d.get(ANNOTATIONS, {}).items():
        if isinstance(values, dict):
            for value in values:
                rv.append(
                    {
                        "type": "Evidence",
                        "label": key,
                        "id": str(value),
                    }
                )
        else:
            rv.append(
                {
                    "type": "Evidence",
                    "label": key,
                    "id": str(values),
                }
            )
    return rv


def from_graphdati(j, use_tqdm: bool = True) -> BELGraph:
    """Convert data from the "normal" network format.

    .. warning:: BioDati crashes when requesting the ``full`` network format, so this isn't yet explicitly supported
    """
    root = j["graph"]
    graph = BELGraph(
        name=root.get("label"),
        version=root["metadata"].get("gd_rev"),
        authors=root["metadata"].get("gd_creator"),
        description=root.get("gd_description"),
    )
    # Just in case you want to find it again
    graph.graph["biodati_network_id"] = root["metadata"]["id"]

    parser = BELParser(
        graph=graph,
        namespace_to_pattern=NAMESPACE_TO_PATTERN,  # To be updated manually depending on what William is up to
    )

    it = root["edges"]
    if use_tqdm:
        it = tqdm(it, desc="iterating edges")

    for i, edge in enumerate(it):
        relation = edge.get("relation")
        if relation is None:
            logger.warning("no relation for edge: %s", edge)

        if relation in {"actsIn", "translocates"}:
            continue  # don't need legacy BEL format

        bel_statement = edge.get("label")  # this is actually the BEL statement
        if bel_statement is None:
            logger.debug("No BEL statement for edge %s", edge)
            continue

        # Fill up that sweet, sweet metadata
        metadata_entries = edge["metadata"]["nanopub_data"]
        for metadata in metadata_entries:
            parser.control_parser.clear()

            citation = metadata["citation_id"]  # as CURIE
            citation_db, citation_id = _parse_biodati_citation(citation)
            if citation_db is None:
                continue
            parser.control_parser.citation_db = citation_db
            parser.control_parser.citation_db_id = citation_id

            # FIXME where is the evidence/support/summary text?
            parser.control_parser.evidence = "No evidence available from BioDai"

            nanopub_id = metadata["nanopub_id"]
            parser.control_parser.annotations["biodati_nanopub_id"] = [nanopub_id]

            annotations = metadata["annotations"]
            parser.control_parser.annotations.update(_parse_biodati_annotations(annotations))

            # Finally, parse the BEL statement (once to go with each set of metadata)
            # TODO change parser to give back pre-compiled info so this doesn't need to be repeated
            try:
                parser.parseString(bel_statement, line_number=i)
            except pyparsing.ParseException as e:
                logger.warning("parse error for %s: %s", bel_statement, e)

    return graph


def _parse_biodati_citation(citation: str) -> Union[Tuple[None, None], Tuple[str, str]]:
    try:
        citation_db, citation_id = citation.split(":")
    except ValueError:
        logger.warning("structured citation not available for %s", citation)
        return None, None

    try:
        citation_db = GRAPHDATI_PUBLICATION_TYPES[citation_db]
    except KeyError:
        logger.warning("invalid citation structure: %s", citation)
        return None, None

    return citation_db, citation_id


def _parse_biodati_annotations(annotations: List[Mapping[str, str]]) -> Mapping[str, Mapping[str, bool]]:
    rv = defaultdict(set)
    for annotation in annotations:
        annotation_curie = annotation["id"]
        annotation_prefix, annotation_id = annotation_curie.split(":", 1)
        rv[annotation_prefix].add(annotation_id)
    return dict(rv)



# -*- coding: utf-8 -*-

"""Fusions.

Gene, RNA, miRNA, and protein  fusions are all represented with the same underlying data structure. Below
it is shown with uppercase letters referring to constants from :code:`pybel.constants` and. For example,
:code:`g(HGNC:BCR, fus(HGNC:JAK2, 1875, 2626))` is represented as:

.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: GENE,
        FUSION: {
            PARTNER_5P: {NAMESPACE: 'HGNC', NAME: 'BCR'},
            PARTNER_3P: {NAMESPACE: 'HGNC', NAME: 'JAK2'},
            RANGE_5P: {
                FUSION_REFERENCE: 'c',
                FUSION_START: '?',
                FUSION_STOP: 1875,
            },
            RANGE_3P: {
                FUSION_REFERENCE: 'c',
                FUSION_START: 2626,
                FUSION_STOP: '?',
            },
        },
    }


.. seealso::

    - BEL 2.0 specification on `fusions (2.6.1)
      <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_fusion_fus>`_
    - PyBEL module :py:class:`pybel.parser.modifiers.get_fusion_language`
    - PyBEL module :py:class:`pybel.parser.modifiers.get_legacy_fusion_language`
"""

from pyparsing import Group, Keyword, Optional, ParserElement, Suppress, oneOf
from pyparsing import pyparsing_common
from pyparsing import pyparsing_common as ppc
from pyparsing import replaceWith

from ..utils import WCW, nest
from ...constants import (
    CONCEPT,
    FUSION,
    FUSION_MISSING,
    FUSION_REFERENCE,
    FUSION_START,
    FUSION_STOP,
    PARTNER_3P,
    PARTNER_5P,
    RANGE_3P,
    RANGE_5P,
)

__all__ = [
    "fusion_tags",
    "get_fusion_language",
    "get_legacy_fusion_langauge",
]

fusion_tags = oneOf(["fus", "fusion"]).setParseAction(replaceWith(FUSION))
reference_seq = oneOf(["r", "p", "c"])
coordinate = pyparsing_common.integer | "?"
missing = Keyword("?")
range_coordinate_unquoted = missing(FUSION_MISSING) | (
    reference_seq(FUSION_REFERENCE) + Suppress(".") + coordinate(FUSION_START) + Suppress("_") + coordinate(FUSION_STOP)
)


def get_fusion_language(concept: ParserElement, permissive: bool = True) -> ParserElement:
    """Build a fusion parser."""
    range_coordinate = Suppress('"') + range_coordinate_unquoted + Suppress('"')

    if permissive:  # permissive to wrong quoting
        range_coordinate = range_coordinate | range_coordinate_unquoted

    return fusion_tags + nest(
        Group(Group(concept)(CONCEPT))(PARTNER_5P),
        Group(range_coordinate)(RANGE_5P),
        Group(Group(concept)(CONCEPT))(PARTNER_3P),
        Group(range_coordinate)(RANGE_3P),
    )


def get_legacy_fusion_langauge(concept: ParserElement, reference: str) -> ParserElement:
    """Build a legacy fusion parser."""
    break_start = (ppc.integer | "?").setParseAction(_fusion_break_handler_wrapper(reference, start=True))
    break_end = (ppc.integer | "?").setParseAction(_fusion_break_handler_wrapper(reference, start=False))

    res = (
        Group(concept(CONCEPT))(PARTNER_5P)
        + WCW
        + fusion_tags
        + nest(
            Group(concept(CONCEPT))(PARTNER_3P)
            + Optional(WCW + Group(break_start)(RANGE_5P) + WCW + Group(break_end)(RANGE_3P)),
        )
    )

    res.setParseAction(_fusion_legacy_handler)
    return res


def _fusion_legacy_handler(_, __, tokens):
    """Handle a legacy fusion."""
    if RANGE_5P not in tokens:
        tokens[RANGE_5P] = {FUSION_MISSING: "?"}
    if RANGE_3P not in tokens:
        tokens[RANGE_3P] = {FUSION_MISSING: "?"}
    return tokens


def _fusion_break_handler_wrapper(reference: str, start: bool):
    def fusion_break_handler(_, __, tokens):
        if tokens[0] == "?":
            tokens[FUSION_MISSING] = "?"
            return tokens
        else:  # The break point is specified as an integer
            tokens[FUSION_REFERENCE] = reference
            tokens[FUSION_START if start else FUSION_STOP] = "?"
            tokens[FUSION_STOP if start else FUSION_START] = int(tokens[0])
            return tokens

    return fusion_break_handler



# -*- coding: utf-8 -*-

"""Test the user API."""

import unittest

import pybel


class TestParse(unittest.TestCase):
    """Test user API."""

    def test_bel_statement(self):
        for bel in [
            "p(hgnc:1234) -> p(hgnc:1235)",
            "p(hgnc:1234) hasVariant p(hgnc:1234, pmod(Ph))",
            "p(hgnc:1) => rxn(reactants(a(chebi:1), a(chebi:2)), products(a(chebi:3), a(chebi:4)))",
            "r(hgnc:1) pos r(hgnc:2)",
            "r(hgnc:1) cor r(hgnc:2)",
            "r(hgnc:1) eq r(ncbigene:5)",
            "p(hgnc:1) -> (p(hgnc:2) -> p(hgnc:3))",
            "p(hgnc:1)",
        ]:
            with self.subTest(bel=bel):
                x = pybel.parse(bel)

    def test_citation(self):
        x = pybel.parse("SET Citation = pmid:1234")

    def test_control(self):
        x = pybel.parse('SET Test = "5"')

    def test_control_multiple(self):
        x = pybel.parse('SET Test = {"1","2","3"}')



# -*- coding: utf-8 -*-

"""Tests for ``pybel.tokens``."""

import unittest

from pybel.constants import ABUNDANCE, CONCEPT, FUNCTION, IDENTIFIER, NAME, NAMESPACE
from pybel.dsl import Abundance
from pybel.testing.utils import n
from pybel.tokens import _simple_po_to_dict


class TestRecover(unittest.TestCase):
    """Test converting dictionaries to DSL."""

    def test_simple(self):
        """Test converting a simple dictionary."""
        namespace, name, identifier = n(), n(), n()

        self.assertEqual(
            Abundance(namespace=namespace, name=name),
            _simple_po_to_dict(
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: namespace,
                        NAME: name,
                    },
                }
            ),
        )

        self.assertEqual(
            Abundance(namespace=namespace, name=name, identifier=identifier),
            _simple_po_to_dict(
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: namespace,
                        NAME: name,
                        IDENTIFIER: identifier,
                    },
                }
            ),
        )

        self.assertEqual(
            Abundance(namespace=namespace, identifier=identifier),
            _simple_po_to_dict(
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: namespace,
                        IDENTIFIER: identifier,
                    },
                }
            ),
        )

        with self.assertRaises(ValueError):
            _simple_po_to_dict(
                {
                    FUNCTION: ABUNDANCE,
                    CONCEPT: {
                        NAMESPACE: namespace,
                    },
                }
            )



# -*- coding: utf-8 -*-

"""Test the manager's citation utilities.

The test data can be created with the following script:

.. code-block:: python

    import json
    from pybel.manager.citation_utils import get_pubmed_citation_response

    DATA = {'29324713', '29359844', '9611787', '25818332', '26438529', '26649137', '27003210'}

    rv = get_pubmed_citation_response(DATA)
    with open('/Users/cthoyt/dev/bel/pybel/tests/test_manager/citation_data.json', 'w') as file:
        json.dump(rv, file, indent=2)

"""

import json
import os
import time
import unittest
from typing import Any, Iterable, Mapping
from unittest import mock

from pybel import BELGraph
from pybel.constants import (
    CITATION,
    CITATION_AUTHORS,
    CITATION_DATE,
    CITATION_JOURNAL,
    CITATION_TYPE_PUBMED,
)
from pybel.dsl import Protein
from pybel.language import CitationDict
from pybel.manager.citation_utils import (
    _enrich_citations,
    enrich_pubmed_citations,
    get_citations_by_pmids,
    sanitize_date,
)
from pybel.manager.models import Citation
from pybel.testing.cases import TemporaryCacheMixin
from pybel.testing.utils import n

HERE = os.path.abspath(os.path.dirname(__file__))

PUBMED_DATA_PATH = os.path.join(HERE, "pubmed_citation_data.json")
with open(PUBMED_DATA_PATH) as _file:
    PUBMED_DATA = json.load(_file)

PMC_DATA_PATH = os.path.join(HERE, "pmc_citation_data.json")
with open(PMC_DATA_PATH) as _file:
    PMC_DATA = json.load(_file)


def _mock_fn(pubmed_identifiers: Iterable[str]) -> Mapping[str, Any]:
    result = {
        "uids": pubmed_identifiers,
    }
    for pmid in pubmed_identifiers:
        result[pmid] = PUBMED_DATA["result"][pmid]
    return {"result": result}


mock_get_pubmed_citation_response = mock.patch(
    "pybel.manager.citation_utils.get_pubmed_citation_response",
    side_effect=_mock_fn,
)


def _mock_get_pmc_csl_item(pmc_id: str) -> Mapping[str, Any]:
    return PMC_DATA[pmc_id]


mock_get_pmc_csl_item = mock.patch(
    "pybel.manager.citation_utils.get_pmc_csl_item",
    side_effect=_mock_get_pmc_csl_item,
)


class TestSanitizeDate(unittest.TestCase):
    """Test sanitization of dates in various formats."""

    def test_sanitize_1(self):
        """Test YYYY Mon DD."""
        self.assertEqual("2012-12-19", sanitize_date("2012 Dec 19"))

    def test_sanitize_2(self):
        """Test YYYY Mon."""
        self.assertEqual("2012-12-01", sanitize_date("2012 Dec"))

    def test_sanitize_3(self):
        """Test YYYY."""
        self.assertEqual("2012-01-01", sanitize_date("2012"))

    def test_sanitize_4(self):
        """Test YYYY Mon-Mon."""
        self.assertEqual("2012-10-01", sanitize_date("2012 Oct-Dec"))

    def test_sanitize_5(self):
        """Test YYYY Season."""
        self.assertEqual("2012-03-01", sanitize_date("2012 Spring"))

    def test_sanitize_6(self):
        """Test YYYY Mon DD-DD."""
        self.assertEqual("2012-12-12", sanitize_date("2012 Dec 12-15"))

    def test_sanitize_7(self):
        """Test YYYY Mon DD-Mon DD."""
        self.assertEqual("2005-01-29", sanitize_date("2005 Jan 29-Feb 4"))

    def test_sanitize_nope(self):
        """Test failure."""
        self.assertEqual(None, sanitize_date("2012 Early Spring"))


class TestPubmed(TemporaryCacheMixin):
    """Tests for citations."""

    def setUp(self):
        super().setUp()
        self.u, self.v = (Protein(n(), n()) for _ in range(2))
        self.pmid = "9611787"
        self.graph = BELGraph()
        self.graph.add_increases(self.u, self.v, citation=self.pmid, evidence=n())

    @mock_get_pubmed_citation_response
    def test_enrich_pubmed(self, *_):
        self.assertEqual(0, self.manager.count_citations())
        get_citations_by_pmids(manager=self.manager, pmids=[self.pmid])
        self.assertEqual(1, self.manager.count_citations())

        c = self.manager.get_citation_by_pmid(self.pmid)
        self.assertIsNotNone(c)
        self.assertIsInstance(c, Citation)
        self.assertEqual(CITATION_TYPE_PUBMED, c.db)
        self.assertEqual(self.pmid, c.db_id)

    @mock_get_pubmed_citation_response
    def test_enrich_pubmed_list(self, *_):
        pmids = [
            "25818332",
            "27003210",
            "26438529",
            "26649137",
        ]

        get_citations_by_pmids(manager=self.manager, pmids=pmids)

        citation = self.manager.get_or_create_citation(namespace=CITATION_TYPE_PUBMED, identifier="25818332")
        self.assertIsNotNone(citation)

    @mock_get_pubmed_citation_response
    def test_enrich_pubmed_list_grouped(self, *_):
        pmids = [
            "25818332",
            "27003210",
            "26438529",
            "26649137",
        ]

        get_citations_by_pmids(manager=self.manager, pmids=pmids, group_size=2)

        citation = self.manager.get_citation_by_pmid("25818332")
        self.assertIsNotNone(citation)

    @mock_get_pubmed_citation_response
    def test_enrich_pubmed_overwrite(self, *_):
        citation = self.manager.get_or_create_citation(namespace=CITATION_TYPE_PUBMED, identifier=self.pmid)
        self.manager.session.commit()
        self.assertIsNone(citation.date)
        self.assertIsNone(citation.title)

        enrich_pubmed_citations(manager=self.manager, graph=self.graph)

        _, _, d = list(self.graph.edges(data=True))[0]
        citation_dict = d[CITATION]
        self.assertIsInstance(citation_dict, CitationDict)

        self.assertIn(CITATION_JOURNAL, citation_dict)
        self.assertIn(CITATION_DATE, citation_dict)
        self.assertEqual("1998-05-01", citation_dict[CITATION_DATE])

        self.assertIn(CITATION_AUTHORS, citation_dict)
        self.assertEqual(
            {"Lewell XQ", "Judd DB", "Watson SP", "Hann MM"},
            set(citation_dict[CITATION_AUTHORS]),
        )

    @mock_get_pubmed_citation_response
    def test_enrich_pubmed_graph(self, *_):
        enrich_pubmed_citations(manager=self.manager, graph=self.graph)

        _, _, d = list(self.graph.edges(data=True))[0]
        citation_dict = d[CITATION]
        self.assertIsInstance(citation_dict, CitationDict)

        self.assertIn(CITATION_JOURNAL, citation_dict)

        self.assertIn(CITATION_DATE, citation_dict)
        self.assertEqual("1998-05-01", citation_dict[CITATION_DATE])

        self.assertIn(CITATION_AUTHORS, citation_dict)
        self.assertEqual(
            {"Lewell XQ", "Judd DB", "Watson SP", "Hann MM"},
            set(citation_dict[CITATION_AUTHORS]),
        )

    @mock_get_pubmed_citation_response
    @unittest.skipIf(os.environ.get("DB") == "mysql", reason="MySQL collation is wonky")
    def test_enrich_pubmed_accent_duplicate(self, *_):
        """Test when two authors, Gomez C and Goméz C are both checked that they are not counted as duplicates."""
        g1 = "Gomez C"
        g2 = "Gómez C"
        pmid_1, pmid_2 = pmids = [
            "29324713",
            "29359844",
        ]
        get_citations_by_pmids(manager=self.manager, pmids=pmids)

        time.sleep(1)

        x = self.manager.get_citation_by_pmid(pmid_1)
        self.assertIsNotNone(x)
        self.assertEqual("Martínez-Guillén JR", x.first.name, msg="wrong first author name")

        self.assertIn(g1, self.manager.object_cache_author)
        self.assertIn(g2, self.manager.object_cache_author)

        a1 = self.manager.get_author_by_name(g1)
        self.assertEqual(g1, a1.name)

        a2 = self.manager.get_author_by_name(g2)
        self.assertEqual(g2, a2.name)


class TestPMC(TemporaryCacheMixin):
    """Tests for citations."""

    def setUp(self):
        super().setUp()
        self.u, self.v = (Protein(n(), n()) for _ in range(2))
        self.citation_identifier = "PMC6611653"
        self.graph = BELGraph()
        self.graph.add_increases(self.u, self.v, citation=("pmc", self.citation_identifier), evidence=n())

    @mock_get_pmc_csl_item
    def test_enrich_pmc(self, *_):
        errors = _enrich_citations(manager=self.manager, graph=self.graph, prefix="pmc")
        self.assertEqual(0, len(errors), msg=f"Got errors: {errors}")
        _, _, d = list(self.graph.edges(data=True))[0]
        citation_dict = d[CITATION]
        self.assertIsInstance(citation_dict, CitationDict)
        self.assertEqual("pmc", citation_dict.namespace)
        self.assertEqual(self.citation_identifier, citation_dict.identifier)

        self.assertIn(CITATION_JOURNAL, citation_dict)
        self.assertEqual("PLoS computational biology", citation_dict[CITATION_JOURNAL])

        self.assertIn(CITATION_DATE, citation_dict)
        self.assertEqual("2019-06-24", citation_dict[CITATION_DATE])

        self.assertIn(CITATION_AUTHORS, citation_dict)
        self.assertLess(0, len(citation_dict[CITATION_AUTHORS]))
        # TODO the eUtils and CSL thing both normalize the way autors look



# -*- coding: utf-8 -*-

"""A parser for BEL.

This module handles parsing BEL relations and validation of semantics.
"""

import itertools as itt
import logging
from functools import lru_cache
from typing import Any, Dict, List, Mapping, Optional, Pattern, Set, Union

import pyparsing
from pyparsing import (
    Group,
    Keyword,
    MatchFirst,
    ParseResults,
    StringEnd,
    Suppress,
    delimitedList,
    oneOf,
    replaceWith,
)

from .baseparser import BaseParser
from .constants import NamespaceTermEncodingMapping
from .modifiers import (
    get_fragment_language,
    get_fusion_language,
    get_gene_modification_language,
    get_gene_substitution_language,
    get_hgvs_language,
    get_legacy_fusion_langauge,
    get_location_language,
    get_protein_modification_language,
    get_protein_substitution_language,
    get_truncation_language,
)
from .parse_concept import ConceptParser
from .parse_control import ControlParser
from .utils import WCW, nest, one_of_tags, triple
from .. import language
from ..constants import (
    ABUNDANCE,
    ACTIVITY,
    ASSOCIATION,
    BINDS,
    BIOPROCESS,
    CAUSES_NO_CHANGE,
    CELL_SECRETION,
    CELL_SURFACE_EXPRESSION,
    COMPLEX,
    COMPOSITE,
    CONCEPT,
    CORRELATION,
    DECREASES,
    DEGRADATION,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    DIRTY,
    EFFECT,
    EQUIVALENT_TO,
    FROM_LOC,
    FUNCTION,
    FUSION,
    GENE,
    IDENTIFIER,
    INCREASES,
    IS_A,
    LINE,
    LOCATION,
    MEMBERS,
    MIRNA,
    MODIFIER,
    NAME,
    NAMESPACE,
    NEGATIVE_CORRELATION,
    NO_CORRELATION,
    PART_OF,
    PATHOLOGY,
    POPULATION,
    POSITIVE_CORRELATION,
    PRODUCTS,
    PROTEIN,
    REACTANTS,
    REACTION,
    REGULATES,
    RELATION,
    RNA,
    SOURCE,
    TARGET,
    TO_LOC,
    TRANSCRIBED_TO,
    TRANSLATED_TO,
    TRANSLOCATION,
    TWO_WAY_RELATIONS,
    VARIANTS,
    belns_encodings,
)
from ..dsl import BaseEntity
from ..exceptions import (
    InvalidEntity,
    InvalidFunctionSemantic,
    MalformedTranslocationWarning,
    MissingAnnotationWarning,
    MissingCitationException,
    MissingSupportWarning,
    NestedRelationWarning,
)
from ..struct.graph import BELGraph
from ..tokens import parse_result_to_dsl

__all__ = [
    "BELParser",
    "modifier_po_to_dict",
    "parse",
]

logger = logging.getLogger("pybel.parser")

###########################
# 2.1 Abundance Functions #
###########################

#: 2.1.1 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xabundancea>
general_abundance_tags = one_of_tags(["a", "abundance"], ABUNDANCE, FUNCTION)

#: 2.1.2 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XcomplexA
complex_tag = one_of_tags(["complex", "complexAbundance"], COMPLEX, FUNCTION)

#: 2.1.3 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XcompositeA
composite_abundance_tag = one_of_tags(["composite", "compositeAbundance"], COMPOSITE, FUNCTION)

#: 2.1.4 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XgeneA
gene_tag = one_of_tags(["g", "geneAbundance"], GENE, FUNCTION)

#: 2.1.5 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XmicroRNAA
mirna_tag = one_of_tags(["m", "microRNAAbundance"], MIRNA, FUNCTION)

#: 2.1.6 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XproteinA
protein_tag = one_of_tags(["p", "proteinAbundance"], PROTEIN, FUNCTION)

#: 2.1.7 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XrnaA
rna_tag = one_of_tags(["r", "rnaAbundance"], RNA, FUNCTION)

######################
# Modifier Functions #
######################

# `2.2.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_protein_modifications>`_
# See below (needs identifier)

#: `2.2.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_variant_var>`_
variant = get_hgvs_language()

#: `2.2.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_proteolytic_fragments>`_
fragment = get_fragment_language()

# `2.2.4 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_cellular_location>`_
# See below (needs identifier)

#: DEPRECATED
#: <http://openbel.org/language/version_1.0/bel_specification_version_1.0.html#_amino_acid_substitutions>
psub = get_protein_substitution_language()

#: DEPRECATED
#: http://openbel.org/language/version_1.0/bel_specification_version_1.0.html#_sequence_variations>
gsub = get_gene_substitution_language()

#: DEPRECATED
#: http://openbel.org/language/version_1.0/bel_specification_version_1.0.html#_truncated_proteins>
trunc = get_truncation_language()

###############################
# 2.3 & 2.4 Process Functions #
###############################

#: 2.3.1 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_biologicalprocess_bp
biological_process_tag = one_of_tags(["bp", "biologicalProcess"], BIOPROCESS, FUNCTION)

#: 2.3.2 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_pathology_path
pathology_tag = one_of_tags(["o", "path", "pathology"], PATHOLOGY, FUNCTION)

population_tag = one_of_tags(["pop", "populationAbundance"], POPULATION, FUNCTION)

#: 2.3.3 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xactivity
activity_tag = one_of_tags(["act", "activity"], ACTIVITY, MODIFIER)

#: 2.4.1 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XmolecularA
molecular_activity_tags = Suppress(oneOf(["ma", "molecularActivity"]))

################################
# 2.5 Transformation Functions #
################################

#: 2.5.1a http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_translocation_tloc
translocation_tag = one_of_tags(["translocation", "tloc"], TRANSLOCATION, MODIFIER)

#: 2.5.1b http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_cellsecretion_sec
cell_secretion_tag = one_of_tags(["sec", "cellSecretion"], CELL_SECRETION, MODIFIER)

#: 2.5.1c http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_cellsurfaceexpression_surf
cell_surface_expression_tag = one_of_tags(["surf", "cellSurfaceExpression"], CELL_SURFACE_EXPRESSION, MODIFIER)

#: 2.5.2 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_degradation_deg
degradation_tags = one_of_tags(["deg", "degradation"], DEGRADATION, MODIFIER)

#: 2.5.3 http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_reaction_rxn
reaction_tags = one_of_tags(["reaction", "rxn"], REACTION, FUNCTION)

#####################
# BEL Relationships #
#####################

#: `3.1.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xincreases>`_
increases_tag = oneOf(["->", "→", "increases"]).setParseAction(replaceWith(INCREASES))

#: `3.1.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XdIncreases>`_
directly_increases_tag = one_of_tags(["=>", "⇒", "directlyIncreases"], DIRECTLY_INCREASES)

#: `3.1.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xdecreases>`_
decreases_tag = one_of_tags(["-|", "decreases"], DECREASES)

#: `3.1.4 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XdDecreases>`_
directly_decreases_tag = one_of_tags(["=|", "directlyDecreases"], DIRECTLY_DECREASES)

#: `3.1.5 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_ratelimitingstepof>`_
rate_limit_tag = Keyword("rateLimitingStepOf")

#: `3.1.6 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xcnc>`_
causes_no_change_tag = one_of_tags(["cnc", "causesNoChange"], CAUSES_NO_CHANGE)

#: `3.1.7 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_regulates_reg>`_
regulates_tag = one_of_tags(["reg", "regulates"], REGULATES)

#: Binds relation
binds_tag = Keyword(BINDS)

#: Correlation relation
correlation_tag = one_of_tags(["cor", "correlation"], CORRELATION)

#: No Correlation relation
no_correlation_tag = one_of_tags(["noCor", "noCorrelation"], NO_CORRELATION)

#: `3.2.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XnegCor>`_
negative_correlation_tag = one_of_tags(["neg", "negativeCorrelation"], NEGATIVE_CORRELATION)

#: `3.2.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XposCor>`_
positive_correlation_tag = one_of_tags(["pos", "positiveCorrelation"], POSITIVE_CORRELATION)

#: `3.2.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xassociation>`_
association_tag = one_of_tags(["--", "association"], ASSOCIATION)

#: `3.3.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_orthologous>`_
orthologous_tag = Keyword("orthologous")

#: `3.3.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_transcribedto>`_
transcribed_tag = oneOf([":>", "transcribedTo"]).setParseAction(replaceWith(TRANSCRIBED_TO))

#: `3.3.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_translatedto>`_
translated_tag = oneOf([">>", "translatedTo"]).setParseAction(replaceWith(TRANSLATED_TO))

#: `3.4.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hasmember>`_
has_member_tag = Keyword("hasMember")

#: `3.4.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hasmembers>`_
has_members_tag = Keyword("hasMembers")

#: `3.4.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hascomponent>`_
has_component_tag = Keyword("hasComponent")

#: `3.4.4 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hascomponents>`_
has_components_tag = Keyword("hasComponents")

#: `3.4.5 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_isa>`_
is_a_tag = Keyword(IS_A)

#: `3.4.6 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_subprocessof>`_
subprocess_of_tag = Keyword("subProcessOf")

#: `3.5.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_analogous>`_
analogous_tag = Keyword("analogousTo")

#: `3.5.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_biomarkerfor>`_
biomarker_tag = Keyword("biomarkerFor")

#: `3.5.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_prognosticbiomarkerfor>`_
prognostic_biomarker_tag = Keyword("prognosticBiomarkerFor")

biomarker_tags = biomarker_tag | prognostic_biomarker_tag

# Computed edges
has_variant_tags = Keyword("hasVariant")
has_reactant_tags = Keyword("hasReactant")
has_product_tags = Keyword("hasProduct")
part_of_reaction_tags = has_reactant_tags | has_product_tags

#: The ``equivalentTp`` relationship has been proposed for BEL 2.0.0+
equivalent_tag = one_of_tags(["eq", EQUIVALENT_TO], EQUIVALENT_TO)

#: The ``partOf`` relationship has been proposed for BEL 2.0.0+
partof_tag = Keyword(PART_OF)


class BELParser(BaseParser):
    """Build a parser backed by a given dictionary of namespaces."""

    def __init__(
        self,
        graph: Optional[BELGraph] = None,
        namespace_to_term_to_encoding: Optional[NamespaceTermEncodingMapping] = None,
        namespace_to_pattern: Optional[Mapping[str, Pattern]] = None,
        annotation_to_term: Optional[Mapping[str, Set[str]]] = None,
        annotation_to_pattern: Optional[Mapping[str, Pattern]] = None,
        annotation_to_local: Optional[Mapping[str, Set[str]]] = None,
        allow_naked_names: bool = False,
        disallow_nested: bool = False,
        disallow_unqualified_translocations: bool = False,
        citation_clearing: bool = True,
        skip_validation: bool = False,
        autostreamline: bool = True,
        required_annotations: Optional[List[str]] = None,
    ) -> None:
        """Build a BEL parser.

        :param graph: The BEL graph to use to store the network
        :param namespace_to_term_to_encoding: A dictionary of {namespace: {name: encoding}}. Delegated to
         :class:`pybel.parser.parse_identifier.IdentifierParser`
        :param namespace_to_pattern: A dictionary of {namespace: compiled regular expression}. Delegated to
         :class:`pybel.parser.parse_identifier.IdentifierParser`
        :param annotation_to_term: A dictionary of {annotation: set of values}. Delegated to
         :class:`pybel.parser.ControlParser`
        :param annotation_to_pattern: A dictionary of {annotation: regular expression strings}. Delegated to
         :class:`pybel.parser.ControlParser`
        :param annotation_to_local: A dictionary of {annotation: set of values}. Delegated to
         :class:`pybel.parser.ControlParser`
        :param allow_naked_names: If true, turn off naked namespace failures. Delegated to
         :class:`pybel.parser.parse_identifier.IdentifierParser`
        :param disallow_nested: If true, turn on nested statement failures. Delegated to
         :class:`pybel.parser.parse_identifier.IdentifierParser`
        :param disallow_unqualified_translocations: If true, allow translocations without TO and FROM clauses.
        :param citation_clearing: Should :code:`SET Citation` statements clear evidence and all annotations?
         Delegated to :class:`pybel.parser.ControlParser`
        :param autostreamline: Should the parser be streamlined on instantiation?
        :param required_annotations: Optional list of required annotations
        """
        self.graph = graph

        self.disallow_nested = disallow_nested
        self.disallow_unqualified_translocations = disallow_unqualified_translocations

        if skip_validation:
            self.control_parser = ControlParser(
                citation_clearing=citation_clearing,
                required_annotations=required_annotations,
            )

            self.concept_parser = ConceptParser(
                allow_naked_names=allow_naked_names,
                skip_validation=skip_validation,
            )
        else:
            self.control_parser = ControlParser(
                annotation_to_term=annotation_to_term,
                annotation_to_pattern=annotation_to_pattern,
                annotation_to_local=annotation_to_local,
                #
                citation_clearing=citation_clearing,
                required_annotations=required_annotations,
            )

            self.concept_parser = ConceptParser(
                namespace_to_term_to_encoding=namespace_to_term_to_encoding,
                namespace_to_pattern=namespace_to_pattern,
                #
                allow_naked_names=allow_naked_names,
                skip_validation=skip_validation,
            )

        self.control_parser.get_line_number = self.get_line_number
        self.concept_parser.get_line_number = self.get_line_number

        concept = Group(self.concept_parser.language)(CONCEPT)

        # 2.2 Abundance Modifier Functions

        #: `2.2.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_protein_modifications>`_
        self.pmod = get_protein_modification_language(
            concept_fqualified=self.concept_parser.identifier_fqualified,
            concept_qualified=self.concept_parser.identifier_qualified,
        )

        #: `2.2.4 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_cellular_location>`_
        self.location = get_location_language(self.concept_parser.language)
        opt_location = pyparsing.Optional(WCW + self.location)

        #: PyBEL BEL Specification variant
        self.gmod = get_gene_modification_language(
            concept_fqualified=self.concept_parser.identifier_fqualified,
            concept_qualified=self.concept_parser.identifier_qualified,
        )

        # 2.6 Other Functions

        #: `2.6.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_fusion_fus>`_
        self.fusion = get_fusion_language(self.concept_parser.language)

        # 2.1 Abundance Functions

        #: `2.1.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XcomplexA>`_
        self.general_abundance = general_abundance_tags + nest(concept + opt_location)

        self.gene_modified = concept + pyparsing.Optional(
            WCW + delimitedList(Group(variant | gsub | self.gmod))(VARIANTS),
        )

        self.gene_fusion = Group(self.fusion)(FUSION)
        self.gene_fusion_legacy = Group(get_legacy_fusion_langauge(concept, "c"))(FUSION)

        #: `2.1.4 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XgeneA>`_
        self.gene = gene_tag + nest(
            MatchFirst(
                [
                    self.gene_fusion,
                    self.gene_fusion_legacy,
                    self.gene_modified,
                ]
            )
            + opt_location,
        )

        self.mirna_modified = (
            concept
            + pyparsing.Optional(
                WCW + delimitedList(Group(variant))(VARIANTS),
            )
            + opt_location
        )

        #: `2.1.5 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XmicroRNAA>`_
        self.mirna = mirna_tag + nest(self.mirna_modified)

        self.protein_modified = concept + pyparsing.Optional(
            WCW
            + delimitedList(Group(MatchFirst([self.pmod, variant, fragment, psub, trunc])))(
                VARIANTS,
            ),
        )

        self.protein_fusion = Group(self.fusion)(FUSION)
        self.protein_fusion_legacy = Group(get_legacy_fusion_langauge(concept, "p"))(FUSION)

        #: `2.1.6 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XproteinA>`_
        self.protein = protein_tag + nest(
            MatchFirst(
                [
                    self.protein_fusion,
                    self.protein_fusion_legacy,
                    self.protein_modified,
                ]
            )
            + opt_location,
        )

        self.rna_modified = concept + pyparsing.Optional(WCW + delimitedList(Group(variant))(VARIANTS))

        self.rna_fusion = Group(self.fusion)(FUSION)
        self.rna_fusion_legacy = Group(get_legacy_fusion_langauge(concept, "r"))(FUSION)

        #: `2.1.7 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XrnaA>`_
        self.rna = rna_tag + nest(
            MatchFirst(
                [
                    self.rna_fusion,
                    self.rna_fusion_legacy,
                    self.rna_modified,
                ]
            )
            + opt_location,
        )

        self.population = population_tag + nest(concept)

        self.single_abundance = MatchFirst(
            [
                self.general_abundance,
                self.gene,
                self.mirna,
                self.protein,
                self.rna,
                self.population,
            ]
        )

        #: `2.1.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XcomplexA>`_
        self.complex_singleton = complex_tag + nest(concept + opt_location)

        self.complex_list = complex_tag + nest(
            delimitedList(Group(self.single_abundance | self.complex_singleton))(MEMBERS) + opt_location,
        )

        self.complex_abundances = self.complex_list | self.complex_singleton

        # Definition of all simple abundances that can be used in a composite abundance
        self.simple_abundance = self.complex_abundances | self.single_abundance
        self.simple_abundance.setParseAction(self.check_function_semantics)

        #: `2.1.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XcompositeA>`_
        self.composite_abundance = composite_abundance_tag + nest(
            delimitedList(Group(self.simple_abundance))(MEMBERS) + opt_location,
        )

        self.abundance = self.simple_abundance | self.composite_abundance

        # 2.4 Process Modifier Function
        # backwards compatibility with BEL v1.0

        molecular_activity_default = oneOf(list(language.activity_labels)).setParseAction(
            handle_molecular_activity_default,
        )

        #: `2.4.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#XmolecularA>`_
        self.molecular_activity = molecular_activity_tags + nest(
            molecular_activity_default | self.concept_parser.language,
        )

        # 2.3 Process Functions

        #: `2.3.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_biologicalprocess_bp>`_
        self.biological_process = biological_process_tag + nest(concept)

        #: `2.3.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_pathology_path>`_
        self.pathology = pathology_tag + nest(concept)

        self.bp_path = self.biological_process | self.pathology
        self.bp_path.setParseAction(self.check_function_semantics)

        self.activity_standard = activity_tag + nest(
            self.simple_abundance + pyparsing.Optional(WCW + Group(self.molecular_activity)(EFFECT)),
        )

        activity_legacy_tags = oneOf(language.activities)(MODIFIER)
        self.activity_legacy = activity_legacy_tags + nest(self.simple_abundance)
        self.activity_legacy.setParseAction(handle_activity_legacy)

        #: `2.3.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#Xactivity>`_
        self.activity = self.activity_standard | self.activity_legacy

        self.process = self.bp_path | self.activity

        # 2.5 Transformation Functions

        from_loc = Suppress(FROM_LOC) + nest(concept(FROM_LOC))
        to_loc = Suppress(TO_LOC) + nest(concept(TO_LOC))

        self.cell_secretion = cell_secretion_tag + nest(self.simple_abundance)
        self.cell_secretion.addParseAction(handle_secretion)

        self.cell_surface_expression = cell_surface_expression_tag + nest(self.simple_abundance)
        self.cell_surface_expression.addParseAction(handle_surface_expression)

        self.translocation_standard = nest(
            self.simple_abundance + WCW + Group(from_loc + WCW + to_loc)(EFFECT),
        )

        self.translocation_legacy = nest(
            self.simple_abundance + WCW + Group(concept(FROM_LOC) + WCW + concept(TO_LOC))(EFFECT),
        )

        self.translocation_legacy.addParseAction(handle_legacy_tloc)
        self.translocation_unqualified = nest(self.simple_abundance)

        if self.disallow_unqualified_translocations:
            self.translocation_unqualified.setParseAction(self.handle_translocation_illegal)

        #: `2.5.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_translocations>`_
        self.translocation = translocation_tag + MatchFirst(
            [
                self.translocation_unqualified,
                self.translocation_standard,
                self.translocation_legacy,
            ]
        )

        #: `2.5.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_degradation_deg>`_
        self.degradation = degradation_tags + nest(self.simple_abundance)

        #: `2.5.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_reaction_rxn>`_
        self.reactants = Suppress(REACTANTS) + nest(delimitedList(Group(self.simple_abundance)))
        self.products = Suppress(PRODUCTS) + nest(delimitedList(Group(self.simple_abundance)))

        self.reaction = reaction_tags + nest(Group(self.reactants)(REACTANTS), Group(self.products)(PRODUCTS))

        self.transformation = MatchFirst(
            [
                self.cell_secretion,
                self.cell_surface_expression,
                self.translocation,
                self.degradation,
                self.reaction,
            ]
        )

        # 3 BEL Relationships

        self.bel_term = MatchFirst([self.transformation, self.process, self.abundance]).streamline()

        self.bel_to_bel_relations = [
            association_tag,
            increases_tag,
            decreases_tag,
            positive_correlation_tag,
            negative_correlation_tag,
            correlation_tag,
            no_correlation_tag,
            binds_tag,
            causes_no_change_tag,
            orthologous_tag,
            is_a_tag,
            equivalent_tag,
            partof_tag,
            directly_increases_tag,
            directly_decreases_tag,
            analogous_tag,
            regulates_tag,
        ]
        self.bel_to_bel = triple(self.bel_term, MatchFirst(self.bel_to_bel_relations), self.bel_term)

        # Mixed Relationships

        #: `3.1.5 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_ratelimitingstepof>`_
        self.rate_limit = triple(
            MatchFirst([self.biological_process, self.activity, self.transformation]),
            rate_limit_tag,
            self.biological_process,
        )

        #: `3.4.6 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_subprocessof>`_
        self.subprocess_of = triple(
            MatchFirst([self.process, self.activity, self.transformation]),
            subprocess_of_tag,
            self.process,
        )

        #: `3.3.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_transcribedto>`_
        self.transcribed = triple(self.gene, transcribed_tag, self.rna)

        #: `3.3.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_translatedto>`_
        self.translated = triple(self.rna, translated_tag, self.protein)

        #: `3.4.1 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hasmember>`_
        self.has_member = triple(self.abundance, has_member_tag, self.abundance)

        #: `3.4.2 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hasmembers>`_
        self.abundance_list = Suppress("list") + nest(delimitedList(Group(self.abundance)))

        self.has_members = triple(self.abundance, has_members_tag, self.abundance_list)
        self.has_members.setParseAction(self.handle_has_members)

        self.has_components = triple(self.abundance, has_components_tag, self.abundance_list)
        self.has_components.setParseAction(self.handle_has_components)

        self.has_list = self.has_members | self.has_components

        # `3.4.3 <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_hascomponent>`_
        self.has_component = triple(
            self.abundance,
            has_component_tag,
            self.abundance,
        )

        self.biomarker = triple(self.bel_term, biomarker_tags, self.process)

        self.has_variant_relation = triple(self.abundance, has_variant_tags, self.abundance)
        self.part_of_reaction = triple(self.reaction, part_of_reaction_tags, self.abundance)

        self.relation = MatchFirst(
            [
                self.bel_to_bel,
                # self.has_member,
                # self.has_component,
                self.subprocess_of,
                self.rate_limit,
                self.biomarker,
                self.transcribed,
                self.translated,
                # self.has_variant_relation,
                # self.part_of_reaction,
            ]
        )
        if self.graph is not None:
            self.relation.setParseAction(self._handle_relation_harness)

        self.inverted_unqualified_relation = MatchFirst(
            [
                self.has_member,
                self.has_component,
            ]
        )
        if self.graph is not None:
            self.inverted_unqualified_relation.setParseAction(self.handle_inverse_unqualified_relation)

        self.normal_unqualified_relation = MatchFirst(
            [
                self.has_member,
                self.has_component,
                self.has_variant_relation,
                self.part_of_reaction,
            ]
        )
        if self.graph is not None:
            self.normal_unqualified_relation.setParseAction(self.handle_unqualified_relation)

        #: 3.1 Causal Relationships - nested.
        causal_relation_tags = MatchFirst(
            [
                increases_tag,
                decreases_tag,
                directly_decreases_tag,
                directly_increases_tag,
            ]
        )

        self.nested_causal_relationship = triple(
            self.bel_term,
            causal_relation_tags,
            nest(triple(self.bel_term, causal_relation_tags, self.bel_term)),
        )

        if self.graph is not None:
            self.nested_causal_relationship.setParseAction(self.handle_nested_relation)

        # has_members is handled differently from all other relations becuase it gets distrinbuted
        self.relation = MatchFirst(
            [
                self.has_list,
                self.nested_causal_relationship,
                self.relation,
                self.inverted_unqualified_relation,
                self.normal_unqualified_relation,
            ]
        )

        self.singleton_term = self.bel_term + StringEnd()
        if self.graph is not None:
            self.singleton_term.setParseAction(self.handle_term)

        self.statement = self.relation | self.singleton_term
        self.language = self.control_parser.language | self.statement
        self.language.setName("BEL")

        super(BELParser, self).__init__(self.language, streamline=autostreamline)

    def parse(self, s: str) -> Mapping[str, Any]:
        """Parse the string."""
        return self.parseString(s).asDict()

    @property
    def _namespace_dict(self) -> Mapping[str, Mapping[str, str]]:
        """Get the dictionary of {namespace: {name: encoding}} stored in the internal identifier parser."""
        return self.concept_parser.namespace_to_name_to_encoding

    @property
    def _allow_naked_names(self) -> bool:
        """Return if naked names should be parsed (``True``), or if errors should be thrown (``False``)."""
        return self.concept_parser.allow_naked_names

    def get_annotations(self) -> Dict:
        """Get the current annotations in this parser."""
        return self.control_parser.get_annotations()

    def clear(self):
        """Clear the graph and all control parser data (current citation, annotations, and statement group)."""
        if self.graph is not None:
            self.graph.clear()
        self.control_parser.clear()

    def handle_nested_relation(self, line: str, position: int, tokens: ParseResults):
        """Handle nested statements.

        If :code:`self.disallow_nested` is True, raises a ``NestedRelationWarning``.

        :raises: NestedRelationWarning
        """
        if self.disallow_nested:
            raise NestedRelationWarning(self.get_line_number(), line, position)

        subject_hash = self._handle_relation_checked(
            line,
            position,
            {
                SOURCE: tokens[SOURCE],
                RELATION: tokens[RELATION],
                TARGET: tokens[TARGET][SOURCE],
            },
        )

        object_hash = self._handle_relation_checked(
            line,
            position,
            {
                SOURCE: tokens[TARGET][SOURCE],
                RELATION: tokens[TARGET][RELATION],
                TARGET: tokens[TARGET][TARGET],
            },
        )
        self.graph.add_transitivity(subject_hash, object_hash)
        return tokens

    def check_function_semantics(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Raise an exception if the function used on the tokens is wrong.

        :raises: InvalidFunctionSemantic
        """
        concept = tokens.get(CONCEPT)
        if not self._namespace_dict or concept is None:
            return tokens

        namespace, name = concept[NAMESPACE], concept[NAME]

        if namespace in self.concept_parser.namespace_to_pattern:
            return tokens

        if self._allow_naked_names and namespace == DIRTY:  # Don't check dirty names in lenient mode
            return tokens

        valid_functions = set(
            itt.chain.from_iterable(
                belns_encodings.get(encoding, set()) for encoding in self._namespace_dict[namespace][name]
            ),
        )

        if not valid_functions:
            raise InvalidEntity(self.get_line_number(), line, position, namespace, name)

        if tokens[FUNCTION] not in valid_functions:
            raise InvalidFunctionSemantic(
                line_number=self.get_line_number(),
                line=line,
                position=position,
                func=tokens[FUNCTION],
                namespace=namespace,
                name=name,
                allowed_functions=valid_functions,
            )

        return tokens

    def handle_term(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle BEL terms (the subject and object of BEL relations)."""
        self.ensure_node(tokens)
        return tokens

    def _handle_list_helper(self, tokens: ParseResults, relation: str) -> ParseResults:
        """Provide the functionality for :meth:`handle_has_members` and :meth:`handle_has_components`."""
        parent_node_dsl = self.ensure_node(tokens[0])

        for child_tokens in tokens[2]:
            child_node_dsl = self.ensure_node(child_tokens)
            # Note that the polarity is switched since this is just for hasMembers
            # and hasComponents, which are both deprecated as of BEL v2.2
            self.graph.add_unqualified_edge(child_node_dsl, parent_node_dsl, relation)

        return tokens

    def handle_has_members(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle list relations like ``p(X) hasMembers list(p(Y), p(Z), ...)``."""
        return self._handle_list_helper(tokens, IS_A)

    def handle_has_components(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle list relations like ``p(X) hasComponents list(p(Y), p(Z), ...)``."""
        return self._handle_list_helper(tokens, PART_OF)

    def _add_qualified_edge_helper(
        self,
        *,
        source,
        source_modifier,
        relation,
        target,
        target_modifier,
        annotations,
    ) -> str:
        """Add a qualified edge from the internal aspects of the parser."""
        m = {
            BINDS: self.graph.add_binds,
        }
        adder = m.get(relation)
        d = dict(
            evidence=self.control_parser.evidence,
            citation=self.control_parser.get_citation(),
            annotations=annotations,
            source_modifier=source_modifier,
            target_modifier=target_modifier,
            **{LINE: self.get_line_number()},
        )
        if adder is not None:
            return adder(source=source, target=target, **d)
        else:
            return self.graph.add_qualified_edge(source=source, target=target, relation=relation, **d)

    def _add_qualified_edge(self, *, source, source_modifier, relation, target, target_modifier) -> str:
        """Add an edge, then adds the opposite direction edge if it should."""
        d = dict(
            relation=relation,
            annotations=self.control_parser.annotations,
        )
        if relation in TWO_WAY_RELATIONS:
            self._add_qualified_edge_helper(
                source=target,
                source_modifier=target_modifier,
                target=source,
                target_modifier=source_modifier,
                **d,
            )
        return self._add_qualified_edge_helper(
            source=source,
            source_modifier=source_modifier,
            target=target,
            target_modifier=target_modifier,
            **d,
        )

    def _handle_relation(self, tokens: ParseResults) -> str:
        """Handle a relation."""
        source = self.ensure_node(tokens[SOURCE])
        source_modifier = modifier_po_to_dict(tokens[SOURCE])
        relation = tokens[RELATION]
        target = self.ensure_node(tokens[TARGET])
        target_modifier = modifier_po_to_dict(tokens[TARGET])

        return self._add_qualified_edge(
            source=source,
            source_modifier=source_modifier,
            relation=relation,
            target=target,
            target_modifier=target_modifier,
        )

    def _handle_relation_harness(self, line: str, position: int, tokens: Union[ParseResults, Dict]) -> ParseResults:
        """Handle BEL relations based on the policy specified on instantiation.

        Note: this can't be changed after instantiation!
        """
        self._handle_relation_checked(line, position, tokens)
        return tokens

    def _handle_relation_checked(self, line, position, tokens):
        if not self.control_parser.citation_is_set:
            raise MissingCitationException(self.get_line_number(), line, position)

        if not self.control_parser.evidence:
            raise MissingSupportWarning(self.get_line_number(), line, position)

        missing_required_annotations = self.control_parser.get_missing_required_annotations()
        if missing_required_annotations:
            raise MissingAnnotationWarning(self.get_line_number(), line, position, missing_required_annotations)

        return self._handle_relation(tokens)

    def handle_unqualified_relation(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle unqualified relations."""
        subject_node_dsl = self.ensure_node(tokens[SOURCE])
        object_node_dsl = self.ensure_node(tokens[TARGET])
        relation = tokens[RELATION]
        self.graph.add_unqualified_edge(subject_node_dsl, object_node_dsl, relation)
        return tokens

    def handle_inverse_unqualified_relation(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle unqualified relations that should go reverse."""
        source = self.ensure_node(tokens[SOURCE])
        target = self.ensure_node(tokens[TARGET])
        relation = tokens[RELATION]
        self.graph.add_unqualified_edge(source=target, target=source, relation=relation)
        return tokens

    def ensure_node(self, tokens: ParseResults) -> BaseEntity:
        """Turn parsed tokens into canonical node name and makes sure its in the graph."""
        node = parse_result_to_dsl(tokens)
        self.graph.add_node_from_data(node)
        return node

    def handle_translocation_illegal(self, line: str, position: int, tokens: ParseResults) -> None:
        """Handle a malformed translocation."""
        raise MalformedTranslocationWarning(self.get_line_number(), line, position, tokens)


# HANDLERS


def handle_molecular_activity_default(_: str, __: int, tokens: ParseResults) -> ParseResults:
    """Handle a BEL 2.0 style molecular activity with BEL default names."""
    upgraded_cls = language.activity_labels[tokens[0]]
    upgraded_concept = language.activity_mapping[upgraded_cls]
    tokens[NAMESPACE] = upgraded_concept.namespace
    tokens[NAME] = upgraded_concept.name
    tokens[IDENTIFIER] = upgraded_concept.identifier
    return tokens


def handle_activity_legacy(_: str, __: int, tokens: ParseResults) -> ParseResults:
    """Handle BEL 1.0 activities."""
    legacy_cls = language.activity_labels[tokens[MODIFIER]]
    tokens[MODIFIER] = ACTIVITY
    tokens[EFFECT] = language.activity_mapping[legacy_cls]
    logger.log(5, "upgraded legacy activity to %s", legacy_cls)
    return tokens


def handle_legacy_tloc(line: str, position: int, tokens: ParseResults) -> ParseResults:
    """Handle translocations that lack the ``fromLoc`` and ``toLoc`` entries."""
    logger.log(5, "legacy translocation statement: %s [%d]", line, position)
    return tokens


def handle_secretion(_, __, tokens: ParseResults) -> ParseResults:
    tokens[MODIFIER] = TRANSLOCATION
    tokens[EFFECT] = {
        FROM_LOC: language.intracellular,
        TO_LOC: language.extracellular,
    }
    return tokens


def handle_surface_expression(_, __, tokens: ParseResults) -> ParseResults:
    tokens[MODIFIER] = TRANSLOCATION
    tokens[EFFECT] = {
        FROM_LOC: language.intracellular,
        TO_LOC: language.cell_surface,
    }
    return tokens


def modifier_po_to_dict(tokens):
    """Get the location, activity, and/or transformation information as a dictionary.

    :return: a dictionary describing the modifier
    :rtype: dict
    """
    attrs = {}

    if LOCATION in tokens:
        attrs[LOCATION] = dict(tokens[LOCATION])

    if MODIFIER not in tokens:
        return attrs

    if tokens[MODIFIER] == DEGRADATION:
        attrs[MODIFIER] = tokens[MODIFIER]

    elif tokens[MODIFIER] == ACTIVITY:
        attrs[MODIFIER] = tokens[MODIFIER]

        if EFFECT in tokens:
            attrs[EFFECT] = dict(tokens[EFFECT])

    elif tokens[MODIFIER] == TRANSLOCATION:
        attrs[MODIFIER] = tokens[MODIFIER]

        if EFFECT in tokens:
            try:
                attrs[EFFECT] = tokens[EFFECT].asDict()
            except AttributeError:  # for when it was auto-upgraded
                attrs[EFFECT] = dict(tokens[EFFECT])

    elif tokens[MODIFIER] == CELL_SECRETION:
        attrs[MODIFIER] = TRANSLOCATION
        attrs[EFFECT] = {
            FROM_LOC: language.intracellular,
            TO_LOC: language.extracellular,
        }

    elif tokens[MODIFIER] == CELL_SURFACE_EXPRESSION:
        attrs[MODIFIER] = TRANSLOCATION
        attrs[EFFECT] = {
            FROM_LOC: language.intracellular,
            TO_LOC: language.cell_surface,
        }

    else:
        raise ValueError("Invalid value for tokens[MODIFIER]: {}".format(tokens[MODIFIER]))

    return attrs


@lru_cache()
def _default_parser():
    return BELParser(skip_validation=True, citation_clearing=False)


@lru_cache()
def parse(s: str, pprint=False):
    """Parse a BEL statement (without validation)."""
    rv = _default_parser().parse(s)
    if pprint:
        import json

        print(json.dumps(rv, indent=2))
    else:
        return rv



# -*- coding: utf-8 -*-

"""Functions for inducing up/downstream causal subgraphs."""

import logging
from typing import Iterable, Union

from .utils import get_subgraph_by_edge_filter
from ...filters.edge_predicate_builders import (
    build_downstream_edge_predicate,
    build_upstream_edge_predicate,
)
from ...pipeline import transformation
from ....dsl import BaseEntity

__all__ = [
    "get_upstream_causal_subgraph",
    "get_downstream_causal_subgraph",
]

logger = logging.getLogger(__name__)


@transformation
def get_upstream_causal_subgraph(graph, nbunch: Union[BaseEntity, Iterable[BaseEntity]]):
    """Induce a sub-graph from all of the upstream causal entities of the nodes in the nbunch.

    :type graph: pybel.BELGraph
    :rtype: pybel.BELGraph
    """
    return get_subgraph_by_edge_filter(graph, build_upstream_edge_predicate(nbunch))


@transformation
def get_downstream_causal_subgraph(graph, nbunch: Union[BaseEntity, Iterable[BaseEntity]]):
    """Induce a sub-graph from all of the downstream causal entities of the nodes in the nbunch.

    :type graph: pybel.BELGraph
    :rtype: pybel.BELGraph
    """
    return get_subgraph_by_edge_filter(graph, build_downstream_edge_predicate(nbunch))



# -*- coding: utf-8 -*-

"""Tests for the internal DSL."""

import unittest

import pybel.constants as pc
from pybel import BELGraph
from pybel.constants import NAME
from pybel.dsl import (
    Abundance,
    ComplexAbundance,
    CompositeAbundance,
    EnumeratedFusionRange,
    Fragment,
    Gene,
    GeneFusion,
    ListAbundanceEmptyException,
    MissingFusionRange,
    Protein,
    Reaction,
    ReactionEmptyException,
)
from pybel.language import Entity
from pybel.testing.utils import n
from pybel.tokens import parse_result_to_dsl
from pybel.utils import ensure_quotes


class TestDSL(unittest.TestCase):
    """Tests for the internal DSL."""

    def test_add_robust_node(self):
        """Test adding a node with both a name and identifier."""
        graph = BELGraph()
        namespace, name, identifier = n(), n(), n()
        node = Protein(namespace=namespace, name=name, identifier=identifier)

        graph.add_node_from_data(node)
        self.assertIn(node, graph)

    def test_add_identified_node(self):
        """Test what happens when a node with only an identifier is added to a graph."""
        graph = BELGraph()
        namespace, identifier = n(), n()
        node = Protein(namespace=namespace, identifier=identifier)
        self.assertNotIn(NAME, node)

        graph.add_node_from_data(node)
        self.assertIn(node, graph)

    def test_add_named_node(self):
        """Test adding a named node to a BEL graph."""
        graph = BELGraph()
        namespace, name = n(), n()
        node = Protein(namespace=namespace, name=name)

        graph.add_node_from_data(node)
        self.assertIn(node, graph)

    def test_missing_information(self):
        """Test that entity and abundance functions raise on missing name/identifier."""
        with self.assertRaises(ValueError):
            Entity(namespace="test")

        with self.assertRaises(ValueError):
            Protein(namespace="test")

        with self.assertRaises(ValueError):
            Protein(namespace="")

        with self.assertRaises(TypeError):
            Protein(namespace="uniprot", name=1234)

        with self.assertRaises(TypeError):
            Protein(namespace="uniprot", identifier=1234)

        with self.assertRaises(ValueError):
            Protein(namespace="uniprot", name="")

        with self.assertRaises(ValueError):
            Protein(namespace="uniprot", identifier="")

        with self.assertRaises(ValueError):
            Protein(namespace="uniprot", identifier="12345", name="")

        with self.assertRaises(ValueError):
            Protein(namespace="uniprot", identifier="", name="123")

    def test_abundance_as_bel_dash_unquoted(self):
        """Test converting an abundance to BEL with a name that needs quotation."""
        namespace, name = "HGNC", "YFG-1"
        node = Abundance(namespace=namespace, name=name)
        self.assertEqual("a(HGNC:YFG-1)", node.as_bel())

    def test_abundance_as_no_quotes(self):
        """Test converting an abundance that doesn't need quotes, but looks crazy."""
        namespace, name = "a-c", "d.e.f"
        node = Abundance(namespace=namespace, name=name)
        self.assertEqual("a(a-c:d.e.f)", node.as_bel())

    def test_abundance_as_bel_quoted(self):
        """Test converting an abundance to BEL with a name that needs quotation."""
        namespace, name = "HGNC", "YFG~1"
        node = Abundance(namespace=namespace, name=name)
        self.assertEqual('a(HGNC:"YFG~1")', node.as_bel())

    def test_abundance_as_bel(self):
        """Test converting an abundance to BEL with a name that does not need quotation."""
        namespace, name = "HGNC", "YFG"
        node = Abundance(namespace=namespace, name=name)
        self.assertEqual("a(HGNC:YFG)", node.as_bel())

    def test_str_has_identifier(self):
        namespace, identifier = n(), n()
        node = Abundance(namespace=namespace, identifier=identifier)
        self.assertEqual(
            "a({namespace}:{identifier})".format(namespace=namespace, identifier=ensure_quotes(identifier)),
            node.as_bel(),
        )

    def test_str_has_both(self):
        namespace, identifier = n(), n()
        node = Abundance(namespace=namespace, identifier=identifier)
        self.assertEqual(
            "a({namespace}:{identifier})".format(namespace=namespace, identifier=ensure_quotes(identifier)),
            node.as_bel(),
        )

    def test_as_tuple(self):
        namespace, name = n(), n()
        node = Abundance(namespace=namespace, name=name)
        self.assertEqual(hash(node), hash(node.as_bel()))

    def test_empty_complex(self):
        """Test that an empty complex causes a failure."""
        with self.assertRaises(ValueError):
            ComplexAbundance(members=[])

    def test_empty_composite(self):
        """Test that an empty complex causes a failure."""
        with self.assertRaises(ValueError):
            CompositeAbundance(members=[])

    def test_complex_with_name(self):
        """Test what happens with a named complex.

        .. code-block::

            complex(SCOMP:"9-1-1 Complex") hasComponent p(HGNC:HUS1)
            complex(SCOMP:"9-1-1 Complex") hasComponent p(HGNC:RAD1)
            complex(SCOMP:"9-1-1 Complex") hasComponent p(HGNC:RAD9A)

        """
        hus1 = Protein(namespace="HGNC", name="HUS1")
        rad1 = Protein(namespace="HGNC", name="RAD1")
        rad9a = Protein(namespace="HGNC", name="RAD9A")
        members = [hus1, rad1, rad9a]

        nine_one_one = ComplexAbundance(members=members, namespace="SCOMP", name="9-1-1 Complex")

        graph = BELGraph()

        graph.add_node_from_data(nine_one_one)
        self.assertIn(nine_one_one, graph)
        self.assertIn(hus1, graph)
        self.assertIn(rad1, graph)
        self.assertIn(rad9a, graph)

    def test_GeneFusion(self):
        """Test serialization of a gene fusion to BEL with a explicit fusion ranges."""
        dsl = GeneFusion(
            Gene("HGNC", "TMPRSS2"),
            Gene("HGNC", "ERG"),
            EnumeratedFusionRange("c", 1, 79),
            EnumeratedFusionRange("c", 312, 5034),
        )
        self.assertEqual('g(fus(HGNC:TMPRSS2, "c.1_79", HGNC:ERG, "c.312_5034"))', dsl.as_bel())

    def test_gene_fusion_missing_implicit(self):
        """Test serialization of a gene fusion to BEL with a implicit missing fusion ranges."""
        dsl = GeneFusion(
            Gene("HGNC", "TMPRSS2"),
            Gene("HGNC", "ERG"),
        )
        self.assertEqual('g(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "?"))', dsl.as_bel())

    def test_gene_fusion_missing_explicit(self):
        """Test serialization of a gene fusion to BEL with an explicit missing fusion ranges."""
        dsl = GeneFusion(
            Gene("HGNC", "TMPRSS2"),
            Gene("HGNC", "ERG"),
            MissingFusionRange(),
            MissingFusionRange(),
        )
        self.assertEqual('g(fus(HGNC:TMPRSS2, "?", HGNC:ERG, "?"))', dsl.as_bel())


class TestCentralDogma(unittest.TestCase):
    """Test functions specific for :class:`CentralDogmaAbundance`s."""

    def test_get_parent(self):
        """Test the get_parent function in :class:`CentralDogmaAbundance`s."""
        ab42 = Protein(name="APP", namespace="HGNC", variants=[Fragment(start=672, stop=713)])
        app = ab42.get_parent()
        self.assertEqual("p(HGNC:APP)", app.as_bel())
        self.assertEqual('p(HGNC:APP, frag("672_713"))', ab42.as_bel())

    def test_with_variants(self):
        """Test the `with_variant` function in :class:`CentralDogmaAbundance`s."""
        app = Protein(name="APP", namespace="HGNC")
        ab42 = app.with_variants(Fragment(start=672, stop=713))
        self.assertEqual("p(HGNC:APP)", app.as_bel())
        self.assertEqual('p(HGNC:APP, frag("672_713"))', ab42.as_bel())

    def test_with_variants_list(self):
        """Test the `with_variant` function in :class:`CentralDogmaAbundance`s."""
        app = Protein(name="APP", namespace="HGNC")
        ab42 = app.with_variants([Fragment(start=672, stop=713)])
        self.assertEqual("p(HGNC:APP)", app.as_bel())
        self.assertEqual('p(HGNC:APP, frag("672_713"))', ab42.as_bel())

    def test_list_abundance_has_contents(self):
        """Test that the construction of list abundance doesn't have empty lists."""
        with self.assertRaises(ListAbundanceEmptyException):
            ComplexAbundance([])

        with self.assertRaises(ListAbundanceEmptyException):
            CompositeAbundance([])

    def test_reaction(self):
        """Add identified reaction."""
        graph = BELGraph()
        reaction = Reaction(
            namespace="rhea",
            identifier="44104",
            reactants=[
                Abundance(namespace="chebi", identifier="17478"),
                Abundance(namespace="chebi", identifier="15377"),
                Abundance(namespace="chebi", identifier="57540"),
            ],
            products=[
                Abundance(namespace="chebi", identifier="29067"),
                Abundance(namespace="chebi", identifier="15378"),
                Abundance(namespace="chebi", identifier="57945"),
            ],
        )
        graph.add_node_from_data(reaction)
        self.assertEqual(7, graph.number_of_nodes())
        self.assertEqual(6, graph.number_of_edges())

    def test_reaction_has_contents(self):
        """Test that the construction of reaction doesn't have empty lists."""
        with self.assertRaises(ReactionEmptyException):
            Reaction([], [])


class TestParse(unittest.TestCase):
    """Test that :func:`parse_result_to_dsl` works correctly."""

    def test_named_complex(self):
        x = ComplexAbundance(
            namespace="a",
            identifier="b",
            members=[
                Protein(namespace="c", identifier="d"),
                Protein(namespace="c", identifier="e"),
            ],
        )

        y = parse_result_to_dsl(dict(x))
        self.assertIsInstance(y, ComplexAbundance)
        self.assertIn(pc.MEMBERS, y)
        self.assertIn(pc.CONCEPT, y)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Entry points for PyKEEN.

PyKEEN is a machine learning library for knowledge graph embeddings that supports node clustering,
link prediction, entity disambiguation, question/answering, and other tasks with knowledge graphs.
It provides an interface for registering plugins using Python's entrypoints under the
``pykeen.triples.extension_importer`` and ``pykeen.triples.prefix_importer`` groups. More specific
information about how the PyBEL plugins are loaded into PyKEEN can be found in PyBEL's
`setup.cfg <https://github.com/pybel/pybel/blob/master/setup.cfg>`_ under the ``[options.entry_points]``
header.

The following example shows how you can parse/load the triples from a BEL document with the `*.bel` extension.

.. code-block:: python

    from urllib.request import urlretrieve
    url = 'https://raw.githubusercontent.com/cthoyt/selventa-knowledge/master/selventa_knowledge/small_corpus.bel'
    urlretrieve(url, 'small_corpus.bel')

    # Example 1A: Make triples factory
    from pykeen.triples import TriplesFactory
    tf = TriplesFactory(path='small_corpus.bel')

    # Example 1B: Use directly in the pipeline, which automatically invokes training/testing set stratification
    from pykeen.pipeline import pipeline
    results = pipeline(
        dataset='small_corpus.bel',
        model='TransE',
    )

The same is true for precompiled BEL documents in the node-link format with the `*.bel.nodelink.json` extension and
the pickle format with the `*.bel.pickle` extension.

The following example shows how you can load/parse the triples from a BEL document stored in BEL Commons using the
``bel-commons`` prefix in combination with the network's identifier.

.. code-block:: python

    # Example 2A: Make a triples factory
    from pykeen.triples import TriplesFactory
    # the network's identifier is 528
    tf = TriplesFactory(path='bel-commons:528')

    # Example 1B: Use directly in the pipeline, which automatically invokes training/testing set stratification
    from pykeen.pipeline import pipeline
    results = pipeline(
        dataset='bel-commons:528',
        model='TransR',
    )

Currently, this relies on the default BEL Commons service provider at https://bel-commons-dev.scai.fraunhofer.de,
whose location might change in the future.
"""

import numpy as np

from .bel_commons_client import from_bel_commons
from .gpickle import from_pickle
from .nodelink import from_nodelink_file
from .triples import to_triples

__all__ = [
    "get_triples_from_bel",
    "get_triples_from_bel_nodelink",
    "get_triples_from_bel_pickle",
    "get_triples_from_bel_commons",
]


def get_triples_from_bel(path: str) -> np.ndarray:
    """Get triples from a BEL file by wrapping :func:`pybel.io.tsv.api.get_triples`.

    :param path: the file path to a BEL Script
    :return: A three column array with head, relation, and tail in each row
    """
    from pybel import from_bel_script

    return _from_bel(path, from_bel_script)


def get_triples_from_bel_nodelink(path: str) -> np.ndarray:
    """Get triples from a BEL Node-link JSON file by wrapping :func:`pybel.io.tsv.api.get_triples`.

    :param path: the file path to a BEL Node-link JSON file
    :return: A three column array with head, relation, and tail in each row
    """
    return _from_bel(path, from_nodelink_file)


def get_triples_from_bel_pickle(path: str) -> np.ndarray:
    """Get triples from a BEL pickle file by wrapping :func:`pybel.io.tsv.api.get_triples`.

    :param path: the file path to a BEL pickle file
    :return: A three column array with head, relation, and tail in each row
    """
    return _from_bel(path, from_pickle)


def get_triples_from_bel_commons(network_id: str) -> np.ndarray:
    """Load a BEL document from BEL Commons by wrapping :func:`pybel.io.tsv.api.get_triples`.

    :param network_id: The network identifier for a graph in BEL Commons
    :return: A three column array with head, relation, and tail in each row
    """
    return _from_bel(str(network_id), from_bel_commons)


def _from_bel(path, bel_importer) -> np.ndarray:
    graph = bel_importer(path)
    triples = to_triples(graph)
    return np.array(triples)



# -*- coding: utf-8 -*-

"""A module for monkey-patching the JSON encoder.

When it's imported. JSONEncoder.default() automatically checks for a special "to_json()"
method and uses it to encode the object if found.

Provided by user martineau at:
http://stackoverflow.com/questions/18478287/making-object-json-serializable-with-regular-encoder/18561055#18561055
"""

from json import JSONEncoder

__all__ = []


def _default(self, obj):
    return getattr(obj.__class__, "to_json", _default.default)(obj)


_default.default = JSONEncoder().default  # Save unmodified default.
JSONEncoder.default = _default  # replacement



# -*- coding: utf-8 -*-

"""Protein Modifications.

The addition of a post-translational modification (PTM) tag results in an entry called 'variants'
in the data dictionary associated with a given node. This entry is a list with dictionaries
describing each of the variants. All variants have the entry 'kind' to identify whether it is
a PTM, gene modification, fragment, or HGVS variant. The 'kind' value for PTM is 'pmod'.

Each PMOD contains an identifier, which is a dictionary with the namespace and name, and can
optionally include the position ('pos') and/or amino acid code ('code').

For example, the node :code:`p(HGNC:GSK3B, pmod(P, S, 9))` is represented with the following:

.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: PROTEIN,
        NAMESPACE: 'HGNC',
        NAME: 'GSK3B',
        VARIANTS: [
            {
                KIND: PMOD,
                IDENTIFIER: {
                    NAMESPACE: BEL_DEFAULT_NAMESPACE
                    NAME: 'Ph',
                },
                PMOD_CODE: 'Ser',
                PMOD_POSITION: 9,
            },
        ],
    }


As an additional example, in :code:`p(HGNC:MAPK1, pmod(Ph, Thr, 202), pmod(Ph, Tyr, 204))`, MAPK is phosphorylated
twice to become active. This results in the following:

.. code::

   {
       FUNCTION: PROTEIN,
       NAMESPACE: 'HGNC',
       NAME: 'MAPK1',
       VARIANTS: [
           {
               KIND: PMOD,
               IDENTIFIER: {
                   NAMESPACE: BEL_DEFAULT_NAMESPACE
                   NAME: 'Ph',

               },
               PMOD_CODE: 'Thr',
               PMOD_POSITION: 202
           },
           {
               KIND: PMOD,
               IDENTIFIER: {
                   NAMESPACE: BEL_DEFAULT_NAMESPACE
                   NAME: 'Ph',

               },
               PMOD_CODE: 'Tyr',
               PMOD_POSITION: 204
           }
       ]
   }

.. seealso::

   - BEL 2.0 specification on `protein modifications
     <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_proteinmodification_pmod>`_
   - PyBEL module :py:class:`pybel.parser.modifiers.get_protein_modification_language`
"""

import logging

from pyparsing import Group, MatchFirst, Optional, ParserElement, ParseResults, oneOf
from pyparsing import pyparsing_common as ppc

from .constants import amino_acid
from ..utils import WCW, nest, one_of_tags
from ...constants import (
    CONCEPT,
    IDENTIFIER,
    KIND,
    NAME,
    NAMESPACE,
    PMOD,
    PMOD_CODE,
    PMOD_POSITION,
)
from ...language import pmod_legacy_labels, pmod_mappings, pmod_namespace

__all__ = [
    "get_protein_modification_language",
]

logger = logging.getLogger(__name__)


def _handle_pmod_default_ns(_, __, tokens: ParseResults) -> ParseResults:
    upgraded = pmod_namespace[tokens[0]]
    return _r(upgraded, tokens)


def _handle_pmod_legacy_ns(line, _, tokens: ParseResults) -> ParseResults:
    upgraded = pmod_legacy_labels[tokens[0]]
    logger.log(5, "legacy pmod() value %s upgraded to %s", line, upgraded)
    return _r(upgraded, tokens)


def _r(upgraded, tokens):
    e = pmod_mappings[upgraded]["xrefs"][0]
    tokens[NAMESPACE] = e.namespace
    tokens[IDENTIFIER] = e.identifier
    tokens[NAME] = e.name
    return tokens


pmod_tag = one_of_tags(tags=["pmod", "proteinModification"], canonical_tag=PMOD, name=KIND)
pmod_default_ns = oneOf(list(pmod_namespace)).setParseAction(_handle_pmod_default_ns)
pmod_legacy_ns = oneOf(list(pmod_legacy_labels)).setParseAction(_handle_pmod_legacy_ns)


def get_protein_modification_language(
    concept_fqualified: ParserElement,
    concept_qualified: ParserElement,
) -> ParserElement:
    """Build a protein modification parser."""
    pmod_concept = MatchFirst(
        [
            concept_fqualified,
            concept_qualified,
            pmod_default_ns,
            pmod_legacy_ns,
        ]
    )

    return pmod_tag + nest(
        Group(pmod_concept)(CONCEPT)
        + Optional(
            WCW + amino_acid(PMOD_CODE) + Optional(WCW + ppc.integer(PMOD_POSITION)),
        ),
    )



# -*- coding: utf-8 -*-

"""Functions for collapsing nodes."""

from . import collapse, protein_rna_origins
from .collapse import *
from .protein_rna_origins import *

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

"""Constants for modifier parsers."""

from pyparsing import Keyword, MatchFirst, oneOf

from ... import language
from ...exceptions import PlaceholderAminoAcidWarning

aa_single = oneOf(list(language.amino_acid_dict.keys()))
aa_single.setParseAction(lambda s, l, t: [language.amino_acid_dict[t[0]]])

aa_triple = oneOf(list(language.amino_acid_dict.values()))

#: In biological literature, the X is used to denote a truncation. Text mining efforts often encode X as an amino
#: acid, for which we will throw an error using :func:`handle_aa_placeholder`
aa_placeholder = Keyword("X")


def handle_aa_placeholder(line, position, tokens):
    """Raise an exception when encountering a placeholder amino acid, ``X``."""
    raise PlaceholderAminoAcidWarning(-1, line, position, tokens[0])


aa_placeholder.setParseAction(handle_aa_placeholder)

amino_acid = MatchFirst([aa_triple, aa_single, aa_placeholder])



# -*- coding: utf-8 -*-

"""This module supports the relation parser by handling statements."""

import logging
import re
from typing import Mapping, Optional, Pattern, Set

from pyparsing import And, MatchFirst, ParseResults, Suppress, Word

from .baseparser import BaseParser
from .constants import NamespaceTermEncodingMapping
from .utils import delimited_quoted_list, ns, qid, quote, word
from ..constants import (
    BEL_KEYWORD_ANNOTATION,
    BEL_KEYWORD_AS,
    BEL_KEYWORD_DEFINE,
    BEL_KEYWORD_DOCUMENT,
    BEL_KEYWORD_LIST,
    BEL_KEYWORD_NAMESPACE,
    BEL_KEYWORD_PATTERN,
    BEL_KEYWORD_SET,
    BEL_KEYWORD_URL,
    DOCUMENT_KEYS,
    METADATA_VERSION,
    belns_encodings,
)
from ..exceptions import (
    InvalidMetadataException,
    RedefinedAnnotationError,
    RedefinedNamespaceError,
    VersionFormatWarning,
)
from ..resources.resources import keyword_to_url
from ..utils import valid_date_version

__all__ = [
    "MetadataParser",
]

logger = logging.getLogger(__name__)

as_tag = Suppress(BEL_KEYWORD_AS)
url_tag = Suppress(BEL_KEYWORD_URL)
list_tag = Suppress(BEL_KEYWORD_LIST)
set_tag = Suppress(BEL_KEYWORD_SET)
define_tag = Suppress(BEL_KEYWORD_DEFINE)

function_tags = Word("".join(belns_encodings))

SEMANTIC_VERSION_STRING_RE = re.compile(
    r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:-(?P<release>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?(?:\+(?P<build>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?",
)

MALFORMED_VERSION_STRING_RE = re.compile(r"(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)?")

NAMESPACE_BLACKLIST = {}  # TODO: {'SCOMP', 'SFAM'}


class MetadataParser(BaseParser):
    """A parser for the document and definitions section of a BEL document.

    .. seealso::

        BEL 1.0 Specification for the `DEFINE <http://openbel.org/language/web/version_1.0/bel_specification_version_1.0.html#_define>`_ keyword
    """

    def __init__(
        self,
        manager,
        namespace_to_term_to_encoding: Optional[NamespaceTermEncodingMapping] = None,
        namespace_to_pattern: Optional[Mapping[str, Pattern]] = None,
        annotation_to_term: Optional[Mapping[str, Set[str]]] = None,
        annotation_to_pattern: Optional[Mapping[str, Pattern]] = None,
        annotation_to_local: Optional[Mapping[str, Set[str]]] = None,
        default_namespace: Optional[Set[str]] = None,
        allow_redefinition: bool = False,
        skip_validation: bool = False,
        upgrade_urls: bool = False,
    ) -> None:
        """Build a metadata parser.

        :param manager: A cache manager
        :param namespace_to_term_to_encoding:
          An enumerated namespace mapping from {namespace keyword: {(identifier, name): encoding}}
        :param namespace_to_pattern:
          A regular expression namespace mapping from {namespace keyword: regex string}
        :param annotation_to_term: Enumerated annotation mapping from {annotation keyword: set of valid values}
        :param annotation_to_pattern: Regular expression annotation mapping from {annotation keyword: regex string}
        :param default_namespace: A set of strings that can be used without a namespace
        :param skip_validation: If true, don't download and cache namespaces/annotations
        """
        #: This metadata parser's internal definition cache manager
        self.manager = manager
        self.disallow_redefinition = not allow_redefinition
        self.skip_validation = skip_validation
        self.upgrade_urls = upgrade_urls

        #: A dictionary of cached {namespace keyword: {(identifier, name): encoding}}
        self.namespace_to_term_to_encoding = namespace_to_term_to_encoding or {}
        #: A set of namespaces's URLs that can't be cached
        self.uncachable_namespaces = set()
        #: A dictionary of {namespace keyword: regular expression string}
        self.namespace_to_pattern = namespace_to_pattern or {}
        #: A set of names that can be used without a namespace
        self.default_namespace = set(default_namespace) if default_namespace is not None else None

        #: A dictionary of cached {annotation keyword: set of values}
        self.annotation_to_term = annotation_to_term or {}
        #: A dictionary of {annotation keyword: regular expression string}
        self.annotation_to_pattern = annotation_to_pattern or {}
        #: A dictionary of cached {annotation keyword: set of values}
        self.annotation_to_local = annotation_to_local or {}

        #: A dictionary containing the document metadata
        self.document_metadata = {}

        #: A dictionary from {namespace keyword: BEL namespace URL}
        self.namespace_url_dict = {}
        #: A dictionary from {annotation keyword: BEL annotation URL}
        self.annotation_url_dict = {}

        self.document = And(
            [
                set_tag,
                Suppress(BEL_KEYWORD_DOCUMENT),
                word("key"),
                Suppress("="),
                qid("value"),
            ]
        )

        namespace_tag = And([define_tag, Suppress(BEL_KEYWORD_NAMESPACE), ns("name"), as_tag])
        self.namespace_url = And([namespace_tag, url_tag, quote("url")])
        self.namespace_pattern = And([namespace_tag, Suppress(BEL_KEYWORD_PATTERN), quote("value")])

        annotation_tag = And([define_tag, Suppress(BEL_KEYWORD_ANNOTATION), ns("name"), as_tag])
        self.annotation_url = And([annotation_tag, url_tag, quote("url")])
        self.annotation_list = And([annotation_tag, list_tag, delimited_quoted_list("values")])
        self.annotation_pattern = And([annotation_tag, Suppress(BEL_KEYWORD_PATTERN), quote("value")])

        self.document.setParseAction(self.handle_document)
        self.namespace_url.setParseAction(self.handle_namespace_url)
        self.namespace_pattern.setParseAction(self.handle_namespace_pattern)
        self.annotation_url.setParseAction(self.handle_annotations_url)
        self.annotation_list.setParseAction(self.handle_annotation_list)
        self.annotation_pattern.setParseAction(self.handle_annotation_pattern)

        self.language = MatchFirst(
            [
                self.document,
                self.namespace_url,
                self.annotation_url,
                self.annotation_list,
                self.annotation_pattern,
                self.namespace_pattern,
            ]
        ).setName("BEL Metadata")

        super().__init__(self.language)

    def handle_document(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle statements like ``SET DOCUMENT X = "Y"``.

        :raises: InvalidMetadataException
        :raises: VersionFormatWarning
        """
        key = tokens["key"]
        value = tokens["value"]

        if key not in DOCUMENT_KEYS:
            raise InvalidMetadataException(self.get_line_number(), line, position, key, value)

        norm_key = DOCUMENT_KEYS[key]

        if norm_key in self.document_metadata:
            logger.warning("Tried to overwrite metadata: %s", key)
            return tokens

        self.document_metadata[norm_key] = value

        if norm_key == METADATA_VERSION:
            self.raise_for_version(line, position, value)

        return tokens

    def raise_for_redefined_namespace(self, line: str, position: int, namespace: str) -> None:
        """Raise an exception if a namespace is already defined.

        :raises: RedefinedNamespaceError
        """
        if self.disallow_redefinition and self.has_namespace(namespace):
            raise RedefinedNamespaceError(self.get_line_number(), line, position, namespace)

    def handle_namespace_url(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle statements like ``DEFINE NAMESPACE X AS URL "Y"``.

        :raises: RedefinedNamespaceError
        :raises: pybel.resources.exc.ResourceError
        """
        namespace_keyword = tokens["name"]

        if namespace_keyword in NAMESPACE_BLACKLIST:
            raise ValueError("Upgrade usage to FamPlex")

        self.raise_for_redefined_namespace(line, position, namespace_keyword)

        url = tokens["url"]
        if self.upgrade_urls and namespace_keyword.lower() in keyword_to_url:
            url = keyword_to_url[namespace_keyword.lower()]

        self.namespace_url_dict[namespace_keyword] = url

        return tokens

    def ensure_resources(self):
        """Load all namespaces/annotations that have been encountered so far during parsing."""
        if self.skip_validation:
            return

        if self.namespace_url_dict:
            keywords, urls = zip(*self.namespace_url_dict.items())
            namespaces = self.manager._ensure_namespace_urls(urls)
            for keyword, namespace in zip(keywords, namespaces):
                self.namespace_to_term_to_encoding[keyword] = namespace.get_term_to_encodings()

        if self.annotation_url_dict:
            keywords, urls = zip(*self.annotation_url_dict.items())
            namespaces = self.manager._ensure_namespace_urls(urls, is_annotation=True)
            for keyword, namespace in zip(keywords, namespaces):
                self.annotation_to_term[keyword] = {entry.name for entry in namespace.entries}

    def handle_namespace_pattern(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle statements like ``DEFINE NAMESPACE X AS PATTERN "Y"``.

        :raises: RedefinedNamespaceError
        """
        namespace = tokens["name"]
        self.raise_for_redefined_namespace(line, position, namespace)
        self.namespace_to_pattern[namespace] = re.compile(tokens["value"])
        return tokens

    def raise_for_redefined_annotation(self, line: str, position: int, annotation: str) -> None:
        """Raise an exception if the given annotation is already defined.

        :raises: RedefinedAnnotationError
        """
        if self.disallow_redefinition and self.has_annotation(annotation):
            raise RedefinedAnnotationError(self.get_line_number(), line, position, annotation)

    def handle_annotations_url(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle statements like ``DEFINE ANNOTATION X AS URL "Y"``.

        :raises: RedefinedAnnotationError
        """
        keyword = tokens["name"]
        self.raise_for_redefined_annotation(line, position, keyword)
        self.annotation_url_dict[keyword] = tokens["url"]
        return tokens

    def handle_annotation_list(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle statements like ``DEFINE ANNOTATION X AS LIST {"Y","Z", ...}``.

        :raises: RedefinedAnnotationError
        """
        annotation = tokens["name"]
        self.raise_for_redefined_annotation(line, position, annotation)
        self.annotation_to_local[annotation] = set(tokens["values"])
        return tokens

    def handle_annotation_pattern(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle statements like ``DEFINE ANNOTATION X AS PATTERN "Y"``.

        :raises: RedefinedAnnotationError
        """
        annotation = tokens["name"]
        self.raise_for_redefined_annotation(line, position, annotation)
        self.annotation_to_pattern[annotation] = re.compile(tokens["value"])
        return tokens

    def has_enumerated_annotation(self, annotation: str) -> bool:
        """Check if this annotation is defined by an enumeration."""
        return annotation in self.annotation_to_term

    def has_regex_annotation(self, annotation: str) -> bool:
        """Check if this annotation is defined by a regular expression."""
        return annotation in self.annotation_to_pattern

    def has_local_annotation(self, annotation: str) -> bool:
        """Check if this annotation is defined by an locally."""
        return annotation in self.annotation_to_local

    def has_annotation(self, annotation: str) -> bool:
        """Check if this annotation is defined."""
        return (
            self.has_enumerated_annotation(annotation)
            or self.has_regex_annotation(annotation)
            or self.has_local_annotation(annotation)
        )

    def has_enumerated_namespace(self, namespace: str) -> bool:
        """Check if this namespace is defined by an enumeration."""
        return namespace in self.namespace_to_term_to_encoding

    def has_regex_namespace(self, namespace: str) -> bool:
        """Check if this namespace is defined by a regular expression."""
        return namespace in self.namespace_to_pattern

    def has_namespace(self, namespace: str) -> bool:
        """Check if this namespace is defined."""
        return self.has_enumerated_namespace(namespace) or self.has_regex_namespace(namespace)

    def raise_for_version(self, line: str, position: int, version: str) -> None:
        """Check that a version string is valid for BEL documents.

        This means it's either in the YYYYMMDD or semantic version format.

        :param line: The line being parsed
        :param position: The position in the line being parsed
        :param str version: A version string
        :raises: VersionFormatWarning
        """
        if valid_date_version(version):
            return

        if not SEMANTIC_VERSION_STRING_RE.match(version):
            raise VersionFormatWarning(self.get_line_number(), line, position, version)



# -*- coding: utf-8 -*-

"""Summary functions for nodes in BEL graphs."""

import itertools as itt
import typing
from collections import Counter, defaultdict
from typing import Any, Iterable, List, Mapping, Optional, Set, Tuple

from ..filters import get_nodes, has_activity, has_variant, is_degraded, is_translocated
from ..graph import BELGraph
from ...constants import (
    ACTIVITY,
    CONCEPT,
    EFFECT,
    FROM_LOC,
    FUSION,
    KIND,
    LOCATION,
    MEMBERS,
    MODIFIER,
    NAME,
    NAMESPACE,
    PARTNER_3P,
    PARTNER_5P,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    TO_LOC,
    TRANSLOCATION,
    VARIANTS,
)
from ...dsl import (
    BaseConcept,
    BaseEntity,
    CentralDogma,
    EntityVariant,
    FusionBase,
    ListAbundance,
    Pathology,
    Reaction,
)
from ...language import Entity

__all__ = [
    "get_functions",
    "count_functions",
    "get_namespaces",
    "count_namespaces",
    "get_unused_namespaces",
    "count_names_by_namespace",
    "get_names",
    "get_names_by_namespace",
    "iterate_node_entities",
    "iterate_entities",
    "node_is_grounded",
    "get_ungrounded_nodes",
    "count_variants",
    "count_pathologies",
    "get_top_pathologies",
    "get_top_hubs",
]


def _function_iterator(graph: BELGraph) -> Iterable[str]:
    """Iterate over the functions in a graph.

    :param graph: A BEL graph
    """
    return (node.function for node in graph)


def get_functions(graph: BELGraph) -> Set[str]:
    """Get the set of all functions used in this graph.

    :param graph: A BEL graph
    :return: A set of functions
    """
    return set(_function_iterator(graph))


def count_functions(graph: BELGraph) -> typing.Counter[str]:
    """Count the frequency of each function present in a graph.

    :param graph: A BEL graph
    :return: A Counter from {function: frequency}
    """
    return Counter(_function_iterator(graph))


def _iterate_namespaces(graph: BELGraph) -> Iterable[str]:
    """Iterate over all namespaces found in the graph.

    :param graph: A BEL graph
    """
    for entity in itt.chain(iterate_entities(graph), _iterate_edge_entities(graph)):
        yield entity.namespace


def _iterate_edge_entities(graph: BELGraph) -> Iterable[Entity]:
    for ((_, _, data), side) in itt.product(graph.edges(data=True), (SOURCE_MODIFIER, TARGET_MODIFIER)):
        side_data = data.get(side)
        if side_data is None:
            continue

        modifier = side_data.get(MODIFIER)
        effect = side_data.get(EFFECT)

        if modifier == ACTIVITY and effect is not None:
            assert isinstance(effect, Entity)
            yield effect
        elif modifier == TRANSLOCATION and effect is not None:
            from_loc = effect[FROM_LOC]
            assert isinstance(from_loc, Entity)
            yield from_loc
            to_loc = effect[TO_LOC]
            assert isinstance(to_loc, Entity)
            yield to_loc

        location = side_data.get(LOCATION)
        if location is not None:
            assert isinstance(location, Entity)
            yield location


def count_namespaces(graph: BELGraph) -> typing.Counter[str]:
    """Count the frequency of each namespace across all nodes (that have namespaces).

    :param graph: A BEL graph
    :return: A Counter from {namespace: frequency}
    """
    return Counter(_iterate_namespaces(graph))


def get_namespaces(graph: BELGraph) -> Set[str]:
    """Get the set of all namespaces used in this graph.

    :param graph: A BEL graph
    :return: A set of namespaces
    """
    return set(_iterate_namespaces(graph))


def get_unused_namespaces(graph: BELGraph) -> Set[str]:
    """Get the set of all namespaces that are defined in a graph, but are never used.

    :param graph: A BEL graph
    :return: A set of namespaces that are included but not used
    """
    return graph.defined_namespace_keywords - get_namespaces(graph)


def get_names(graph: BELGraph) -> Mapping[str, Set[str]]:
    """Get all names for each namespace.

    :param graph: A BEL graph
    """
    rv = defaultdict(set)
    for namespace, name in _identifier_filtered_iterator(graph):
        rv[namespace].add(name)
    return dict(rv)


def iterate_entities(graph: BELGraph) -> Iterable[Entity]:
    """Iterate over all entities in the graph.

    :param graph: A BEL graph
    """
    for node in graph:
        yield from iterate_node_entities(node)


def iterate_node_entities(node: BaseEntity) -> Iterable[Entity]:
    """Iterate over all named entities that comprise a node.

    This includes the node's name, the members/reactants/products of the node,
    the fusion partners, the named variants, and all recursive ones too.

    :param node: A BEL node

    Entities in a simple protein:

    >>> from pybel.dsl import Protein
    >>> from pybel.language import Entity
    >>> from pybel.struct.summary import iterate_entities
    >>> protein = Protein(namespace='hgnc', identifier='1455', name='CALR')
    >>> protein_entities = list(iterate_node_entities(protein))
    >>> assert [Entity(namespace='hgnc', identifier='1455', name='CALR')] == protein_entities

    Entities in a protein complex:

    >>> from pybel.dsl import Protein, ComplexAbundance
    >>> from pybel.language import Entity
    >>> from pybel.struct.summary import iterate_entities
    >>> protein_1 = Protein(namespace='hgnc', identifier='1')
    >>> protein_2 = Protein(namespace='hgnc', identifier='2')
    >>> complex_1 = ComplexAbundance([protein_1, protein_2])
    >>> complex_entities = list(iterate_node_entities(complex_1))
    >>> assert [Entity(namespace='hgnc', identifier='1'), Entity(namespace='hgnc', identifier='2')] == complex_entities
    """
    if isinstance(node, BaseConcept):
        yield node.entity
    if isinstance(node, ListAbundance):
        for member in node.members:
            yield from iterate_node_entities(member)
    if isinstance(node, Reaction):
        for member in itt.chain(node.reactants, node.products):
            yield from iterate_node_entities(member)
    if isinstance(node, CentralDogma):
        for variant in node.variants or []:
            if isinstance(variant, EntityVariant):
                yield variant.entity
    if isinstance(node, FusionBase):
        yield from iterate_node_entities(node.partner_5p)
        yield from iterate_node_entities(node.partner_3p)


def _identifier_filtered_iterator(graph) -> Iterable[Tuple[str, str]]:
    """Iterate over names in the given namespace."""
    for data in graph:
        for pair in _get_node_names(data):
            yield pair

        for member in data.get(MEMBERS, []):
            for pair in _get_node_names(member):
                yield pair

    for ((_, _, data), side) in itt.product(graph.edges(data=True), (SOURCE_MODIFIER, TARGET_MODIFIER)):
        side_data = data.get(side)
        if side_data is None:
            continue

        modifier = side_data.get(MODIFIER)
        effect = side_data.get(EFFECT)

        if modifier == ACTIVITY and effect is not None and NAMESPACE in effect and NAME in effect:
            yield effect[NAMESPACE], effect[NAME]

        elif modifier == TRANSLOCATION and effect is not None:
            from_loc = effect.get(FROM_LOC)
            if NAMESPACE in from_loc and NAME in from_loc:
                yield from_loc[NAMESPACE], from_loc[NAME]

            to_loc = effect.get(TO_LOC)
            if NAMESPACE in to_loc and NAME in to_loc:
                yield to_loc[NAMESPACE], to_loc[NAME]

        location = side_data.get(LOCATION)
        if location is not None and NAMESPACE in location and NAME in location:
            yield location[NAMESPACE], location[NAME]


def _get_node_names(data: Mapping[str, Any]) -> Iterable[Tuple[str, str]]:
    if CONCEPT in data:
        yield data[CONCEPT][NAMESPACE], data[CONCEPT][NAME]

    elif FUSION in data:
        partner_5p_concept = data[FUSION][PARTNER_5P][CONCEPT]
        partner_3p_concept = data[FUSION][PARTNER_3P][CONCEPT]
        yield partner_5p_concept[NAMESPACE], partner_5p_concept[NAME]
        yield partner_3p_concept[NAMESPACE], partner_3p_concept[NAME]

    if VARIANTS in data:
        for variant in data[VARIANTS]:
            concept = variant.get(CONCEPT)
            if concept is not None and NAMESPACE in concept and NAME in concept:
                yield concept[NAMESPACE], concept[NAME]


def _namespace_filtered_iterator(graph: BELGraph, namespace: str) -> Iterable[str]:
    """Iterate over names in the given namespace.

    :param graph: A BEL graph
    :param namespace: A namespace prefix
    """
    for it_namespace, name in _identifier_filtered_iterator(graph):
        if namespace == it_namespace:
            yield name


def count_names_by_namespace(graph: BELGraph, namespace: str) -> typing.Counter[str]:
    """Get the set of all of the names in a given namespace that are in the graph.

    :param graph: A BEL graph
    :param namespace: A namespace prefix
    :return: A counter from {name: frequency}

    :raises IndexError: if the namespace is not defined in the graph.
    """
    if namespace not in graph.defined_namespace_keywords:
        raise IndexError("{} is not defined in {}".format(namespace, graph))

    return Counter(_namespace_filtered_iterator(graph, namespace))


def get_names_by_namespace(graph: BELGraph, namespace: str) -> Set[str]:
    """Get the set of all of the names in a given namespace that are in the graph.

    :param pybel.BELGraph graph: A BEL graph
    :param namespace: A namespace prefix
    :return: A set of names belonging to the given namespace that are in the given graph

    :raises IndexError: if the namespace is not defined in the graph.
    """
    if namespace not in graph.defined_namespace_keywords:
        raise IndexError("{} is not defined in {}".format(namespace, graph))

    return set(_namespace_filtered_iterator(graph, namespace))


def count_variants(graph: BELGraph) -> typing.Counter[str]:
    """Count how many of each type of variant a graph has.

    :param graph: A BEL graph
    """
    return Counter(variant_data[KIND] for data in graph if has_variant(graph, data) for variant_data in data[VARIANTS])


def get_top_hubs(graph: BELGraph, *, n: Optional[int] = 15) -> List[Tuple[BaseEntity, int]]:
    """Get the top hubs in the graph by BEL.

    :param graph: A BEL graph
    :param n: The number of top hubs to return. If None, returns all nodes
    """
    return Counter(dict(graph.degree())).most_common(n=n)


def count_pathologies(graph: BELGraph) -> typing.Counter[BaseEntity]:
    """Count the number of edges in which each pathology is incident.

    :param graph: A BEL graph
    """
    # Don't double count relationships
    edges = {tuple(sorted([u, v], key=lambda node: node.as_bel())) for u, v in graph.edges()}
    return Counter(node for node in itt.chain.from_iterable(edges) if isinstance(node, Pathology))


def get_top_pathologies(graph: BELGraph, n: Optional[int] = 15) -> List[Tuple[BaseEntity, int]]:
    """Get the top highest relationship-having edges in the graph by BEL.

    :param graph: A BEL graph
    :param n: The number of top connected pathologies to return. If None, returns all nodes
    """
    return count_pathologies(graph).most_common(n)


def get_ungrounded_nodes(graph: BELGraph) -> Set[BaseEntity]:
    """Get all ungrounded nodes in the graph.

    :param graph: A BEL graph
    """
    return {node for node in graph if not node_is_grounded(node)}


def node_is_grounded(node: BaseEntity) -> bool:
    """Check if a node is grounded.

    :param node: A BEL node
    """
    return all(entity.identifier is not None and entity.name is not None for entity in iterate_node_entities(node))


def get_degradations(graph: BELGraph) -> Set[BaseEntity]:
    """Get all nodes that are degraded."""
    return get_nodes(graph, is_degraded)


def get_activities(graph: BELGraph) -> Set[BaseEntity]:
    """Get all nodes that have molecular activities."""
    return get_nodes(graph, has_activity)


def get_translocated(graph: BELGraph) -> Set[BaseEntity]:
    """Get all nodes that are translocated."""
    return get_nodes(graph, is_translocated)


def count_modifications(graph: BELGraph) -> Counter:
    """Get a modifications count dictionary."""
    return Counter(
        remove_falsy_values(
            {
                "Translocations": len(get_translocated(graph)),
                "Degradations": len(get_degradations(graph)),
                "Molecular Activities": len(get_activities(graph)),
            }
        )
    )


def remove_falsy_values(counter: Mapping[Any, int]) -> Mapping[Any, int]:
    """Remove all values that are zero."""
    return {label: count for label, count in counter.items() if count}



# -*- coding: utf-8 -*-

"""Transport functions for Amazon Web Services (AWS).

AWS has a cloud-based file storage service called S3 that can be programatically
accessed using the :mod:`boto3` package. This module provides functions for quickly
wrapping upload/download of BEL graphs using the gzipped Node-Link schema.
"""

import logging
from io import BytesIO
from typing import Any, Optional

from .nodelink import from_nodelink_gz_io, to_nodelink_gz_io
from ..struct import BELGraph

__all__ = [
    "to_s3",
    "from_s3",
]

logger = logging.getLogger(__name__)
S3Client = Any


def to_s3(graph: BELGraph, *, bucket: str, key: str, client: Optional[S3Client] = None) -> None:
    """Save BEL to S3 as gzipped node-link JSON.

    If you don't specify an instantiated client, PyBEL will do its best to load a default
    one using :func:`boto3.client` like in the following example:

    .. code-block:: python

        import pybel
        from pybel.examples import sialic_acid_graph

        graph = pybel.to_s3(
            sialic_acid_graph,
            bucket='your bucket',
            key='your file name.bel.nodelink.json.gz',
        )

    However, if you would like to configure your own, you can do it with something like this:

    .. code-block:: python

        import boto3
        s3_client = boto3.client('s3')

        import pybel
        from pybel.examples import sialic_acid_graph

        graph = pybel.to_s3(
            sialic_acid_graph,
            client=s3_client,
            bucket='your bucket',
            key='your file name.bel.nodelink.json.gz',
        )

    .. warning:: This assumes you already have credentials set up on your machine

    If you don't already have a bucket, you can create one using ``boto3`` by following
    this tutorial: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-example-creating-buckets.html
    """
    if client is None:
        import boto3

        client = boto3.client("s3")
    io = to_nodelink_gz_io(graph)
    client.upload_fileobj(io, bucket, key)


def from_s3(*, bucket: str, key: str, client: Optional[S3Client] = None) -> BELGraph:
    """Get BEL from gzipped node-link JSON from Amazon S3.

    If you don't specify an instantiated client, PyBEL will do its best to load a default
    one using :func:`boto3.client` like in the following example:

    .. code-block:: python

        graph = pybel.from_s3(bucket='your bucket', key='your file name.bel.nodelink.json.gz')

    However, if you would like to configure your own, you can do it with something like this:

    .. code-block:: python

        import boto3
        s3_client = boto3.client('s3')

        import pybel
        graph = pybel.from_s3(
            client=s3_client,
            bucket='your bucket',
            key='your file name.bel.nodelink.json.gz',
        )
    """
    if client is None:
        import boto3

        client = boto3.client("s3")
    io = BytesIO()
    client.download_fileobj(bucket, key, io)
    io.seek(0)
    return from_nodelink_gz_io(io)



# -*- coding: utf-8 -*-

"""Constants for PyBEL tests."""

import logging
import re
import unittest
from json import dumps

from pybel import BELGraph
from pybel.canonicalize import edge_to_bel
from pybel.constants import (
    ANNOTATIONS,
    ASSOCIATION,
    CITATION,
    CITATION_TYPE_PUBMED,
    DECREASES,
    DIRECTLY_DECREASES,
    EVIDENCE,
    IDENTIFIER,
    INCREASES,
    LINE,
    METADATA_AUTHORS,
    METADATA_DESCRIPTION,
    METADATA_LICENSES,
    METADATA_NAME,
    METADATA_VERSION,
    NAMESPACE,
    PART_OF,
    RELATION,
)
from pybel.dsl import BaseEntity, ComplexAbundance, Pathology, Protein
from pybel.dsl.namespaces import hgnc
from pybel.exceptions import (
    BELParserWarning,
    BELSyntaxError,
    CitationTooShortException,
    IllegalAnnotationValueWarning,
    InvalidCitationLengthException,
    InvalidCitationType,
    InvalidFunctionSemantic,
    InvalidPubMedIdentifierWarning,
    MalformedTranslocationWarning,
    MissingAnnotationKeyWarning,
    MissingAnnotationRegexWarning,
    MissingCitationException,
    MissingMetadataException,
    MissingNamespaceNameWarning,
    MissingNamespaceRegexWarning,
    MissingSupportWarning,
    NakedNameWarning,
    NestedRelationWarning,
    PlaceholderAminoAcidWarning,
    UndefinedAnnotationWarning,
    UndefinedNamespaceWarning,
    VersionFormatWarning,
)
from pybel.language import Entity, citation_dict
from pybel.parser.parse_bel import BELParser
from pybel.parser.parse_control import ControlParser
from pybel.testing.constants import test_bel_thorough
from pybel.utils import subdict_matches
from tests.constant_helper import (
    BEL_THOROUGH_EDGES,
    BEL_THOROUGH_NODES,
    citation_1,
    evidence_1,
    expected_test_simple_metadata,
    expected_test_thorough_metadata,
)

logger = logging.getLogger(__name__)

OPENBEL_DOMAIN = "http://resources.openbel.org"
OPENBEL_ANNOTATION_RESOURCES = OPENBEL_DOMAIN + "/belframework/20150611/annotation/"

test_citation_dict = citation_dict(namespace=CITATION_TYPE_PUBMED, identifier="1235813")

SET_CITATION_TEST = f'SET Citation = {{"{test_citation_dict.namespace}", "{test_citation_dict.identifier}"}}'
test_evidence_text = "I read it on Twitter"
test_set_evidence = f'SET Evidence = "{test_evidence_text}"'

HGNC_KEYWORD = "HGNC"
MESH_DISEASES_KEYWORD = "MeSHDisease"
MESH_DISEASES_URL = OPENBEL_ANNOTATION_RESOURCES + "mesh-diseases.belanno"

akt1 = hgnc(name="AKT1")
egfr = hgnc(name="EGFR")
fadd = hgnc(name="FADD")
casp8 = hgnc(name="CASP8")


def update_provenance(control_parser: ControlParser) -> None:
    """Put a default evidence and citation in a BEL parser."""
    control_parser.citation_db = test_citation_dict.namespace
    control_parser.citation_db_id = test_citation_dict.identifier
    control_parser.evidence = test_evidence_text


def assert_has_node(self: unittest.TestCase, node: BaseEntity, graph: BELGraph, **kwargs):
    """Check if a node with the given properties is contained within a graph."""
    self.assertIsInstance(node, BaseEntity)

    self.assertIn(
        node,
        graph,
        msg="{} not found in graph. Other nodes:\n{}".format(node.as_bel(), "\n".join(n.as_bel() for n in graph)),
    )

    if kwargs:
        missing = set(kwargs) - set(graph.nodes[node])
        self.assertFalse(missing, msg="Missing {} in node data".format(", ".join(sorted(missing))))
        self.assertTrue(
            all(kwarg in graph.nodes[node] for kwarg in kwargs),
            msg="Missing kwarg in node data",
        )
        self.assertEqual(
            kwargs,
            {k: graph.nodes[node][k] for k in kwargs},
            msg="Wrong values in node data",
        )


def any_dict_matches(dict_of_dicts, query_dict) -> bool:
    """

    :param dict_of_dicts:
    :param query_dict:
    :return:
    """
    return any(query_dict == sd for sd in dict_of_dicts.values())


def any_subdict_matches(dict_of_dicts, query_dict) -> bool:
    """Checks if dictionary target_dict matches one of the subdictionaries of a

    :param dict[any,dict] dict_of_dicts: dictionary of dictionaries
    :param dict query_dict: dictionary
    :return: if dictionary target_dict matches one of the subdictionaries of a
    """
    return any(subdict_matches(sub_dict, query_dict) for sub_dict in dict_of_dicts.values())


def _remove_line(d):
    return {k: v for k, v in d.items() if k != LINE}


def assert_has_edge(
    self: unittest.TestCase,
    u: BaseEntity,
    v: BaseEntity,
    graph: BELGraph,
    *,
    only: bool = False,
    permissive: bool = True,
    use_identifiers: bool = False,
    **expected_edge_data,
):
    """A helper function for checking if an edge with the given properties is contained within a graph."""
    self.assertIsInstance(u, BaseEntity)
    self.assertIsInstance(v, BaseEntity)

    self.assertTrue(
        graph.has_edge(u, v),
        msg="Edge ({}, {}) not in graph. Other edges:\n{}".format(
            u,
            v,
            "\n".join(edge_to_bel(u, v, d, use_identifiers=use_identifiers) for u, v, d in graph.edges(data=True)),
        ),
    )

    if not expected_edge_data:
        return

    if ANNOTATIONS in expected_edge_data:
        expected_edge_data[ANNOTATIONS] = graph._clean_annotations(expected_edge_data[ANNOTATIONS])

    if only:
        _key, actual_edge_data = list(graph[u][v].items())[0]
        self.assertEqual(
            _remove_line(expected_edge_data),
            _remove_line(actual_edge_data),
            msg="Only entry not equal",
        )
    else:
        actual_dicts = {k: _remove_line(v) for k, v in graph[u][v].items()}
        if permissive:
            matches = any_subdict_matches(actual_dicts, _remove_line(expected_edge_data))
        else:
            matches = any_dict_matches(actual_dicts, _remove_line(expected_edge_data))

        msg = "No edge ({}, {}) with correct properties. expected:\n {}\nbut got:\n{}".format(
            u,
            v,
            dumps(expected_edge_data, indent=2, sort_keys=True),
            dumps(actual_dicts, indent=2, sort_keys=True),
        )
        self.assertTrue(matches, msg=msg)


class TestGraphMixin(unittest.TestCase):
    """A test case with additional functions for testing graphs."""

    def assert_has_node(self, g: BELGraph, n: BaseEntity, **kwargs):
        """Help assert node membership."""
        assert_has_node(self, n, g, **kwargs)

    def assert_has_edge(
        self,
        g: BELGraph,
        u: BaseEntity,
        v: BaseEntity,
        only=False,
        permissive=True,
        **kwargs,
    ):
        """Help assert edge membership."""
        assert_has_edge(self, u, v, g, only=only, permissive=permissive, **kwargs)


class TestTokenParserBase(unittest.TestCase):
    """A test case that has a BEL parser available."""

    @classmethod
    def setUpClass(cls):
        """Build a BEL graph and BEL parser that persist through the class."""
        cls.parser = BELParser(
            BELGraph(),  # gets overwritten in each test
            autostreamline=False,
            disallow_unqualified_translocations=True,
            namespace_to_pattern={
                "HGNC": re.compile(r".*"),
                "SNP": re.compile(r".*"),
                "CHEBI": re.compile(r".*"),
                "REF": re.compile(r".*"),
                "MOD": re.compile(r".*"),
                "EFO": re.compile(r".*"),
                "GO": re.compile(r".*"),
                "UBERON": re.compile(r".*"),
                "FPLX": re.compile(r".*"),
                "MESH": re.compile(r".*"),
                "MGI": re.compile(r".*"),
                "dbSNP": re.compile(r".*"),
            },
        )

    @property
    def graph(self):
        return self.parser.graph

    def setUp(self):
        """Clear the parser at the beginning of each test."""
        self.parser.clear()
        self.parser.graph = BELGraph()  # because all the good stuff in the graph is cleared

    def assert_has_node(self, member: BaseEntity, **kwargs):
        """Assert that this test case's graph has the given node."""
        assert_has_node(self, member, self.graph, **kwargs)

    def assert_has_edge(self, u: BaseEntity, v: BaseEntity, only: bool = False, **kwargs):
        """Assert that this test case's graph has the given edge."""
        assert_has_edge(self, u, v, self.graph, only=only, **kwargs)

    def add_default_provenance(self):
        """Add a default citation and evidence to the parser."""
        update_provenance(self.parser.control_parser)


def help_check_hgnc(test_case: unittest.TestCase, namespace_dict) -> None:
    """Assert that the namespace dictionary is correct."""
    test_case.assertIn(HGNC_KEYWORD, namespace_dict)

    mhs2 = "7071", "MHS2"
    test_case.assertIn(mhs2, namespace_dict[HGNC_KEYWORD])
    test_case.assertEqual(set("G"), set(namespace_dict[HGNC_KEYWORD][mhs2]))

    miatnb = "50731", "MIATNB"
    test_case.assertIn(miatnb, namespace_dict[HGNC_KEYWORD])
    test_case.assertEqual(set("GR"), set(namespace_dict[HGNC_KEYWORD][miatnb]))

    mia = "7076", "MIA"
    test_case.assertIn(mia, namespace_dict[HGNC_KEYWORD])
    test_case.assertEqual(set("GRP"), set(namespace_dict[HGNC_KEYWORD][mia]))


class BelReconstitutionMixin(TestGraphMixin):
    """A test case that has checks for properly loading several BEL Scripts."""

    def bel_simple_reconstituted(self, graph: BELGraph, check_metadata: bool = True):
        """Check that test_bel.bel was loaded properly."""
        self.assertIsNotNone(graph)
        self.assertIsInstance(graph, BELGraph)

        if check_metadata:
            self.assertIsNotNone(graph.document)
            self.assertEqual(expected_test_simple_metadata[METADATA_NAME], graph.name)
            self.assertEqual(expected_test_simple_metadata[METADATA_VERSION], graph.version)

        self.assertEqual(4, graph.number_of_nodes())

        # FIXME this should work, but is getting 8 for the upgrade function
        # self.assertEqual(6, graph.number_of_edges(),
        #                  msg='Edges:\n{}'.format('\n'.join(map(str, graph.edges(keys=True, data=True)))))

        for node in graph:
            self.assertIsInstance(node, BaseEntity)

        self.assertIn(akt1, graph)
        self.assertIn(egfr, graph)
        self.assertIn(fadd, graph)
        self.assertIn(casp8, graph)

        bel_simple_citation_1 = citation_dict(namespace=CITATION_TYPE_PUBMED, identifier="123455")
        bel_simple_citation_2 = citation_dict(namespace=CITATION_TYPE_PUBMED, identifier="123456")

        evidence_1_extra = "Evidence 1 w extra notes"
        evidence_2 = "Evidence 2"
        evidence_3 = "Evidence 3"

        assert_has_edge(
            self,
            akt1,
            egfr,
            graph,
            only=True,
            **{
                RELATION: INCREASES,
                CITATION: bel_simple_citation_1,
                EVIDENCE: evidence_1_extra,
                ANNOTATIONS: {
                    "Species": [Entity(namespace="Species", identifier="9606")],
                },
            },
        )
        assert_has_edge(
            self,
            egfr,
            fadd,
            graph,
            only=True,
            **{
                RELATION: DECREASES,
                ANNOTATIONS: {
                    "Species": [Entity(namespace="Species", identifier="9606")],
                    "CellLine": [Entity(namespace="CellLine", identifier="10B9 cell")],
                },
                CITATION: bel_simple_citation_1,
                EVIDENCE: evidence_2,
            },
        )
        assert_has_edge(
            self,
            egfr,
            casp8,
            graph,
            only=True,
            **{
                RELATION: DIRECTLY_DECREASES,
                ANNOTATIONS: {
                    "Species": [Entity(namespace="Species", identifier="9606")],
                    "CellLine": [Entity(namespace="CellLine", identifier="10B9 cell")],
                },
                CITATION: bel_simple_citation_1,
                EVIDENCE: evidence_2,
            },
        )
        assert_has_edge(
            self,
            fadd,
            casp8,
            graph,
            only=True,
            **{
                RELATION: INCREASES,
                ANNOTATIONS: {
                    "Species": [Entity(namespace="Species", identifier="10116")],
                },
                CITATION: bel_simple_citation_2,
                EVIDENCE: evidence_3,
            },
        )
        assert_has_edge(
            self,
            akt1,
            casp8,
            graph,
            only=True,
            **{
                RELATION: ASSOCIATION,
                ANNOTATIONS: {
                    "Species": [Entity(namespace="Species", identifier="10116")],
                },
                CITATION: bel_simple_citation_2,
                EVIDENCE: evidence_3,
            },
        )
        assert_has_edge(
            self,
            casp8,
            akt1,
            graph,
            only=True,
            **{
                RELATION: ASSOCIATION,
                ANNOTATIONS: {
                    "Species": [Entity(namespace="Species", identifier="10116")],
                },
                CITATION: bel_simple_citation_2,
                EVIDENCE: evidence_3,
            },
        )

    def bel_thorough_reconstituted(
        self,
        graph: BELGraph,
        check_metadata: bool = True,
        check_warnings: bool = True,
        check_provenance: bool = True,
        check_citation_name: bool = True,
        check_path: bool = True,
    ):
        """Check that thorough.bel was loaded properly.

        :param graph: A BEL graph
        :param check_metadata: Check the graph's document section is correct
        :param check_warnings: Check the graph produced the expected warnings
        :param check_provenance: Check the graph's definition section is correct
        :param check_citation_name: Check that the names in the citations get reconstituted. This isn't strictly
                                         necessary since this data can be looked up
        :param check_path: Check the graph contains provenance for its original file
        """
        self.assertIsNotNone(graph)
        self.assertIsInstance(graph, BELGraph)

        if check_warnings:
            self.assertEqual(
                0,
                len(graph.warnings),
                msg="Document warnings:\n{}".format("\n".join(map(str, graph.warnings))),
            )

        if check_metadata:
            self.assertLessEqual(set(expected_test_thorough_metadata), set(graph.document))
            self.assertEqual(expected_test_thorough_metadata[METADATA_NAME], graph.name)
            self.assertEqual(expected_test_thorough_metadata[METADATA_VERSION], graph.version)
            self.assertEqual(expected_test_thorough_metadata[METADATA_DESCRIPTION], graph.description)

        if check_path:
            self.assertEqual(test_bel_thorough, graph.path)

        if check_provenance:
            self.assertEqual({"CHEBI", "HGNC", "GO", "MESHD", "TESTNS2"}, set(graph.namespace_url))
            self.assertEqual({"dbSNP"}, set(graph.namespace_pattern))
            self.assertIn("TESTAN1", graph.annotation_list)
            self.assertIn("TESTAN2", graph.annotation_list)
            self.assertEqual(
                2,
                len(graph.annotation_list),
                msg="Wrong number of locally defined annotations",
            )
            self.assertEqual({"TestRegex"}, set(graph.annotation_pattern))

        for node in graph:
            self.assertIsInstance(node, BaseEntity)

        self.assertEqual(
            sorted(BEL_THOROUGH_NODES, key=str),
            sorted(graph, key=str),
            msg="Nodes not equal",
        )

        # FIXME
        # self.assertEqual(set((u, v) for u, v, _ in e), set(g.edges()))

        self.assertLess(0, graph.number_of_edges())

        for u, v, data in BEL_THOROUGH_EDGES:
            assert_has_edge(self, u, v, graph, permissive=True, **data)

    def bel_slushy_reconstituted(self, graph: BELGraph, check_metadata: bool = True, check_warnings: bool = True):
        """Check that slushy.bel was loaded properly."""
        self.assertIsNotNone(graph)
        self.assertIsInstance(graph, BELGraph)

        if check_metadata:
            self.assertIsNotNone(graph.document)
            self.assertIsInstance(graph.document, dict)

            expected_test_slushy_metadata = {
                METADATA_NAME: "Worst. BEL Document. Ever.",
                METADATA_DESCRIPTION: "This document outlines all of the evil and awful work that is possible during BEL curation",
                METADATA_VERSION: "0.0",
                METADATA_AUTHORS: "Charles Tapley Hoyt",
                METADATA_LICENSES: "WTF License",
            }

            self.assertEqual(expected_test_slushy_metadata[METADATA_NAME], graph.name)
            self.assertEqual(expected_test_slushy_metadata[METADATA_VERSION], graph.version)
            self.assertEqual(expected_test_slushy_metadata[METADATA_DESCRIPTION], graph.description)

        if check_warnings:
            expected_warnings = [
                (0, MissingMetadataException),
                (3, VersionFormatWarning),
                (26, MissingAnnotationKeyWarning),
                (29, MissingAnnotationKeyWarning),
                (34, CitationTooShortException),
                (37, InvalidCitationType),
                (40, InvalidPubMedIdentifierWarning),
                (43, MissingCitationException),
                (48, MissingAnnotationKeyWarning),
                (51, MissingAnnotationKeyWarning),
                (54, MissingSupportWarning),
                (59, NakedNameWarning),
                (62, UndefinedNamespaceWarning),
                (65, MissingNamespaceNameWarning),
                (68, UndefinedAnnotationWarning),
                (71, MissingAnnotationKeyWarning),
                (74, IllegalAnnotationValueWarning),
                (77, MissingAnnotationRegexWarning),
                (80, MissingNamespaceRegexWarning),
                (83, MalformedTranslocationWarning),
                (86, PlaceholderAminoAcidWarning),
                (89, NestedRelationWarning),
                (92, InvalidFunctionSemantic),
                # (95, Exception),
                (98, BELSyntaxError),
            ]

            for warning_tuple in graph.warnings:
                self.assertEqual(
                    3,
                    len(warning_tuple),
                    msg="Warning tuple is wrong size: {}".format(warning_tuple),
                )

            _sliced_warnings = [(w.line_number, w.__class__) for _, w, _ in graph.warnings]
            self.assertEqual(expected_warnings, _sliced_warnings, msg="wrong warnings")

            for (el, ew), (_, exc, _) in zip(expected_warnings, graph.warnings):
                self.assertIsInstance(exc, BELParserWarning)
                self.assertIsInstance(exc.line, str)
                self.assertIsInstance(exc.line_number, int)
                self.assertIsInstance(exc.position, int)
                self.assertEqual(
                    el,
                    exc.line_number,
                    msg="Expected {} on line {} but got {} on line {}: {}".format(
                        ew, el, exc.__class__, exc.line_number, exc
                    ),
                )
                self.assertIsInstance(exc, ew, msg="Line: {}".format(el))

        for node in graph:
            self.assertIsInstance(node, BaseEntity)

        self.assertIn(akt1, graph, msg="AKT1 not in graph")
        self.assertIn(egfr, graph, msg="EGFR not in graph")

        self.assertEqual(
            2,
            graph.number_of_nodes(),
            msg="Wrong number of nodes retrieved from slushy.bel",
        )
        self.assertEqual(
            1,
            graph.number_of_edges(),
            msg="Wrong nunber of edges retrieved from slushy.bel",
        )

        assert_has_edge(
            self,
            akt1,
            egfr,
            graph,
            only=True,
            **{
                RELATION: INCREASES,
                CITATION: citation_1,
                EVIDENCE: evidence_1,
            },
        )

    def bel_isolated_reconstituted(self, graph: BELGraph):
        """Run the isolated node test."""
        self.assertIsNotNone(graph)
        self.assertIsInstance(graph, BELGraph)

        adgrb1 = Protein(namespace="HGNC", name="ADGRB1")
        adgrb2 = Protein(namespace="HGNC", name="ADGRB2")
        adgrb_complex = ComplexAbundance([adgrb1, adgrb2])
        achlorhydria = Pathology(namespace="MESHD", name="Achlorhydria")

        for node in graph:
            self.assertIsInstance(node, BaseEntity)

        self.assertIn(adgrb1, graph)
        self.assertIn(adgrb2, graph)
        self.assertIn(adgrb_complex, graph)
        self.assertIn(achlorhydria, graph)

        assert_has_edge(self, adgrb1, adgrb_complex, graph, only=True, relation=PART_OF)
        assert_has_edge(self, adgrb2, adgrb_complex, graph, only=True, relation=PART_OF)



# -*- coding: utf-8 -*-

"""Output functions for BEL graphs to Neo4j."""

from tqdm.autonotebook import tqdm

from ..constants import (
    ANNOTATIONS,
    CITATION,
    EVIDENCE,
    FUSION,
    MEMBERS,
    NAMESPACE,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    VARIANTS,
)
from ..utils import flatten_dict

__all__ = [
    "to_neo4j",
]


def to_neo4j(graph, neo_connection, use_tqdm: bool = False):
    """Upload a BEL graph to a Neo4j graph database using :mod:`py2neo`.

    :param pybel.BELGraph graph: A BEL Graph
    :param neo_connection: A :mod:`py2neo` connection object. Refer to the
     `py2neo documentation <http://py2neo.org/v3/database.html#the-graph>`_ for how to build this object.
    :type neo_connection: str or py2neo.Graph

    Example Usage:

    >>> import py2neo
    >>> import pybel
    >>> from pybel.examples import sialic_acid_graph
    >>> neo_graph = py2neo.Graph("http://localhost:7474/db/data/")  # use your own connection settings
    >>> pybel.to_neo4j(sialic_acid_graph, neo_graph)
    """
    import py2neo

    if isinstance(neo_connection, str):
        neo_connection = py2neo.Graph(neo_connection)

    tx = neo_connection.begin()

    node_map = {}

    nodes = list(graph)
    if use_tqdm:
        nodes = tqdm(nodes, desc="nodes")

    for node in nodes:
        if NAMESPACE not in node or VARIANTS in node or MEMBERS in node or FUSION in node:
            attrs = {"name": node.as_bel()}
        else:
            attrs = {"namespace": node.namespace}

            if node.name and node.identifier:
                attrs["name"] = node.name
                attrs["identifier"] = node.identifier
            elif node.identifier and not node.name:
                attrs["name"] = node.identifier
            elif node.name and not node.identifier:
                attrs["name"] = node.name

        node_map[node] = py2neo.Node(node.function, **attrs)

        tx.create(node_map[node])

    edges = graph.edges(keys=True, data=True)
    if use_tqdm:
        edges = tqdm(edges, desc="edges")

    for u, v, key, node in edges:
        rel_type = node[RELATION]

        d = node.copy()
        del d[RELATION]

        attrs = {}

        annotations = d.pop(ANNOTATIONS, None)
        if annotations:
            for annotation, values in annotations.items():
                attrs[annotation] = list(values)

        citation = d.pop(CITATION, None)
        if citation:
            attrs[CITATION] = citation.curie

        if EVIDENCE in d:
            attrs[EVIDENCE] = d[EVIDENCE]

        for side in (SOURCE_MODIFIER, TARGET_MODIFIER):
            side_data = d.get(side)
            if side_data:
                attrs.update(flatten_dict(side_data, parent_key=side))

        rel = py2neo.Relationship(node_map[u], rel_type, node_map[v], key=key, **attrs)
        tx.create(rel)

    tx.commit()



# -*- coding: utf-8 -*-

"""Tests for getters."""

import unittest

from pybel import BELGraph
from pybel.dsl import ComplexAbundance, Protein, Rna
from pybel.struct.getters import get_tf_pairs
from pybel.testing.utils import n


def _tf_up(graph, protein, rna):
    graph.add_directly_increases(
        ComplexAbundance([protein, rna.get_gene()]),
        rna,
        citation=n(),
        evidence=n(),
    )


def _tf_down(graph, protein, rna):
    graph.add_directly_decreases(
        ComplexAbundance([protein, rna.get_gene()]),
        rna,
        citation=n(),
        evidence=n(),
    )


def _bel_pair_key(k):
    return tuple(map(str, k))


class TestGetters(unittest.TestCase):
    """Tests for getters."""

    def test_get_tf_pairs(self):
        """Test iterating over transcription factor pairs."""
        graph = BELGraph()
        p1, p2, p3 = (Protein("test", str(i)) for i in range(1, 4))
        r4, r5, r6 = (Rna("test", str(j)) for j in range(4, 7))

        g4 = r4.get_gene()
        self.assertIsNotNone(g4)
        g5 = r5.get_gene()
        self.assertIsNotNone(g5)

        c14, c25 = ComplexAbundance([p1, g4]), ComplexAbundance([p2, g5])
        _tf_up(graph, p1, r4)
        _tf_down(graph, p2, r5)
        graph.add_correlation(p3, r6, citation=n(), evidence=n())

        self.assertEqual({p1, p2, p3, r4, r5, r6, g4, g5, c14, c25}, set(graph))

        expected_edges = [
            (c14, r4),
            (p1, c14),
            (g4, c14),
            (c25, r5),
            (p2, c25),
            (g5, c25),
            (p3, r6),
            (r6, p3),
        ]
        sorted_expected_edges = sorted(expected_edges, key=_bel_pair_key)
        sorted_actual_edges = sorted(graph.edges(), key=_bel_pair_key)

        self.assertEqual(sorted_expected_edges, sorted_actual_edges)

        pairs = set(get_tf_pairs(graph))
        expected_pairs = {(p1, r4, +1), (p2, r5, -1)}
        self.assertEqual(expected_pairs, pairs)
